""" module to compute wall equilibrium """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np


def compute_equilibrium(hot_t_ad, cold_t_ad, hot_h, cold_h, metal, ceram,
                        ep_metal, ep_ceram):
    """ compute the wall equilibrium

            - - - - - - - >
            HOT SIDE  hot_h, hot_t_ad

         ^ phi
        _|___________________  t_ceram_hot
         |       Layer 2
        _|___________________ t_eq
         |      Layer 1
        _|___________________ t_metal_hot
         |
            COLD SIDE cold_h, cold_t_ad
            - - - - >

    """
    t_eq = np.copy(cold_t_ad)
    t_eq_m1 = np.copy(cold_t_ad)

    def estimate_phi_and_temps(t_ceram, t_metal):
        """ return phi and temperatures """
        r_th_ceram = ceram.thermal_resistance(ep_ceram, t_ceram)
        r_th_metal = metal.thermal_resistance(ep_metal, t_metal)

        phi = (hot_t_ad - cold_t_ad) / (
                1.0 / hot_h
                + r_th_ceram
                + r_th_metal
                + 1.0 / cold_h)

        t_ceram_hot = hot_t_ad - phi / hot_h
        t_ceram_metal = t_ceram_hot - phi * r_th_ceram
        t_metal_cold = t_ceram_metal - phi * r_th_metal

        return t_ceram_hot, t_ceram_metal, t_metal_cold

    t_ceram_hot, t_eq, t_metal_cold = estimate_phi_and_temps(t_ceram=t_eq,
                                                             t_metal=t_eq)

    # Stop convergence at 1K
    while np.max(np.abs(t_eq - t_eq_m1)) > 1.0:
        t_ceram_hot, t_eq, t_metal_cold = estimate_phi_and_temps(
            t_ceram=0.5 * (t_ceram_hot + t_eq),
            t_metal=0.5 * (t_metal_cold + t_eq))
        return t_ceram_hot, t_eq, t_metal_cold



""" Module to test the heat transfer
resolution on grids curvilinear 2D grids """

import arnica.utils.mesh_tool as msh
import scipy.sparse as scp
import scipy.sparse.linalg as linalg
from arnica.solvers_2d.core_fd import Metrics2d
import arnica.solvers_2d.boundary as bndy
from arnica.utils.nparray2xmf import NpArray2Xmf
import numpy as np
import arnica.utils.mesh_tool as msh


def compute_constants(params):
    """ compute_constants parameters

    Parameters :
    ------------
    params : dict of setup parameters


    Returns :
    ------------
    const : dict of constant parameters
    """

    const = {}
    const["h_ratio"] = params["h_cold"] / params["h_hot"]
    const["biot"] = params["h_hot"] * params["width"] / params["lambda"]
    const["t_ref"] = (params["t_cold"]
                      + ((params["t_hot"] - params["t_cold"])
                         * (1. / (1. + const["h_ratio"]))))
    const["dx"] = params["typ_size"] / params["res"]

    const["lam_ov_rhocp"] = params["lambda"] / (params["rho"] * params["cp"])

    const["dt"] = params["fourier"] * const["dx"] ** 2 / const["lam_ov_rhocp"]
    const["total_time"] = params["typ_size"] ** 2 / const["lam_ov_rhocp"]
    const["exp_iterations"] = const["total_time"] / const["dt"]

    for key in const:
        print(key
              + ":"
              + str(const[key]))

    return const


def slope_function(coor, min_val, max_val):
    """
    create a slope x test function, with analytically known derivatives

    Parameters :
    -----------
    coor : numpy array (n,m) , coordinates
    min_val : minimum value
    max_val ; maximum value

    Returns :
    ---------
    slop : numpy array (n,m) , slope test function

    """

    min_c = coor.min()
    max_c = coor.max()

    out = np.ones_like(coor) * min_val
    out += (coor - min_c) / (max_c - min_c) * (max_val - min_val)

    return out


def build_heat_equation(metric):
    """
    Create the LHS matrix corresponding to heat equation
    Parameters
    ----------
    metric : an instance of the class Metrics2d containing the gradient operators

    Returns
    -------
    LHS matrix
    RHS matrix
    """

    # Left Hand Side
    grad_x = metric.grad_x_csr
    grad_y = metric.grad_y_csr
    laplacian = grad_x.dot(grad_x) + grad_y.dot(grad_y)

    lhs = laplacian

    # Right-Hand Side
    rhs = np.zeros(metric.shp1d)

    return lhs, rhs


def set_bc_heat_transfer(bc_params):
    """
    Fill the generic bc_values from wall_temperature.
    (bc_dirichlet function is generic, we have to tell the solver that the
    value that we want to impose is the wall temperature here)

    Parameters
    ----------
    bc_params: dictionary of boundary parameters

    Returns
    -------
    bc_params: updated dictionary of boundary parameters
    """

    for boundary_name, bc_dict in bc_params.items():

        if bc_dict["type"] == 'Dirichlet':
            bc_dict["bc_values"] = bc_dict["Values"]["wall_temperature"]

    return bc_params


def main_2d_condution(params):
    """ Set up main 2D heat transfer resolution

    Parameters :
    ------------
    params : dict of setup parameters

    """

    print("\n\tmain ---> Create mesh")
    x_coor, y_coor = msh.get_mesh(params["mesh"])

    if params['mesh']["dilate_grid"]:
        x_coor, y_coor, _ = msh.dilate_center(x_coor, y_coor)

    print("\n\tmain ---> Define Boundaries")
    bnd = bndy.Boundary2d(params["boundaries"])

    print("\n\tmain ---> Compute Metrics")
    metric = Metrics2d(x_coor,
                       y_coor,
                       periodic_ns=bnd.periodic_ns,
                       periodic_we=bnd.periodic_we)

    print("\n\tmain --> Build LHS and RHS")
    lhs_csr, rhs_csr = build_heat_equation(metric)

    temp_east = params["boundaries"]["East"]["Values"]["wall_temperature"]
    temp_west = params["boundaries"]["West"]["Values"]["wall_temperature"]
    temp_final = slope_function(x_coor, temp_west, temp_east)

    temp0_1d = np.ones(metric.shp1d) * temp_east
    # temp0_2d = temp0_1d.reshape(metric.shp2d)

    if params["solver"]["method"] == "iterative":
        # mat_bc_csr, rhs = apply_bc_csr(metric.lapl, metric.size_i,
        #                                metric.size_j,
        #                                params["t_cold"], params["t_hot"])

        params["boundaries"] = set_bc_heat_transfer(params["boundaries"])

        lhs_csr_bc, rhs_csr_bc = bndy.apply_bc(lhs_csr, rhs_csr, metric,
                                               params["boundaries"])

        temp1d, info = scp.linalg.bicgstab(lhs_csr_bc, rhs_csr_bc, x0=temp0_1d)

        print(">> Result bicg ", info)
        temp = temp1d.reshape(metric.shp2d)

    # if params["method"] == "direct":
    #     const = compute_constants(params)
    #     temp = explicit_solve(
    #         temp0_2d,
    #         metric,
    #         params,
    #         const,
    #         temp_final)

    def gap(temp):
        """ return the maximum absolute error """
        return np.abs(temp_final - temp).max()

    print("Max error:", gap(temp))

    outh5 = NpArray2Xmf("out.h5")
    outh5.create_grid(x_coor, y_coor, np.zeros(metric.shp2d))
    outh5.add_field(temp, "temp")
    outh5.add_field(temp_final, "temp_analytical")
    outh5.add_field(temp_final - temp, "temp_error")
    outh5.dump()


if __name__ == "__main__":
    # Properties
    temp_west = 600.
    h_west = 3000.

    temp_east = 1500.
    h_east = 1500.

    # Mesh definition
    mesh_rect = {"kind": 'rect',
                 "x_max": 1.,
                 "y_max": 1.,
                 "x_res": 51.,
                 "y_res": 50.,
                 "dilate_grid": False}

    # Boundary conditions
    # Default
    boundary = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    boundary["West"]["type"] = "Dirichlet"
    boundary["West"]["Values"] = {"wall_temperature": temp_west,
                                  "heat_transfer_coeff": h_west}

    boundary["East"]["type"] = "Dirichlet"
    boundary["East"]["Values"] = {"wall_temperature": temp_east,
                                  "heat_transfer_coeff": h_east}

    field = {}
    field['rho'] = 7900.
    field['cp'] = 435.
    field['lambda'] = 11.4

    # Solver options
    solver = {"method": "iterative",
              # "method": "direct",
              "tol": 1.0,
              "fourier": 0.8}

    # Create parameter dictionary
    params = {}
    params["boundaries"] = boundary
    params["mesh"] = mesh_rect
    params["field"] = field
    params["solver"] = solver

    main_2d_condution(params)



import matplotlib.pyplot as plt
import numpy as np
from arnica.solvers_2d.radiation import filter_stupid_characters
from matplotlib.colors import LogNorm


def heat_map_mesh(x, y, z, show=False, save=True):
    """ heat map plot of skin """

    plt.rcParams['image.cmap'] = 'Blues'

    def get_bins(x, y):
        """Computes equal bins from aspect ratio"""
        l_y = np.abs(np.max(y) - np.min(x))
        l_x = np.abs(np.max(x) - np.min(x))
        ratio = l_y / l_x
        return 500., 500. * ratio

    def plot_skin(tuple_data, tuple_labels):
        """Plot skin as hist2d"""
        x, y = tuple_data
        x_label, y_label = tuple_labels

        fig = plt.figure()
        ax = fig.add_subplot(111)

        title = "%s, %s" % tuple_labels
        ax.set_title(title)
        ax.hist2d(x, y, bins=get_bins(x, y), norm=LogNorm())
        ax.set_xlabel("%s" % x_label)
        ax.set_ylabel("%s" % y_label)
        ax.set_aspect(aspect=1)
        if save:
            name = filter_stupid_characters(title)
            plt.savefig('%s' % name)

    radius = np.hypot(y, z)
    plot_skin((x, radius), ('$x$', '$r$'))
    plot_skin((x, z), ('$y$', '$z$'))
    plot_skin((x, y), ('$x$', '$y$'))
    plot_skin((z, y), ('$z$', '$y$'))

    if show:
        plt.show()


def scatter_plot_mesh(x, y, z, axisym=False, show=False):
    """ scatter plot of skin """

    if axisym:
        radius = np.sqrt(np.square(x) + np.square(y))
        theta = np.arctan2(z, y)

        plt.figure()
        plt.title(r"x, theta")
        plt.scatter(x, theta, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("theta")

        plt.figure()
        plt.title("x, r")
        plt.scatter(x, radius, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("x")

        plt.figure()
        plt.title("r, theta")
        plt.scatter(radius, theta, edgecolors='none', alpha=0.25)
        plt.xlabel("r")
        plt.xlabel("theta")
    else:
        plt.figure()
        plt.title("z, y")
        plt.scatter(z, y, edgecolors='none', alpha=0.25)
        plt.xlabel("z")
        plt.xlabel("y")

        plt.figure()
        plt.title("x, y")
        plt.scatter(x, y, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("y")

        plt.figure()
        plt.title("x, z")
        plt.scatter(x, z, edgecolors='none', alpha=0.25)
        plt.xlabel("x")
        plt.xlabel("z")

    if show:
        plt.show()



""" Module to test the p1 (radiation) solver for grey gas """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import arnica.utils.mesh_tool as msh
import numpy as np
import scipy.sparse as scp
import scipy.sparse.linalg as linalg
from arnica.solvers_2d.core_fd import Metrics2d
import arnica.solvers_2d.boundary as bndy
from arnica.utils.nparray2xmf import NpArray2Xmf

def filter_stupid_characters(string):
    for char in ["$", "\\", " ", "__", "___"]:
        string = string.replace(char, "_")

    if string.startswith("_"):
        string = string[1:-1]

    return string

def get_p1_eq_field(metric, absorption_coeff, spectral_radiance):
    """
    Create the LHS matrix corresponding to P1 equation

    Parameters
    ----------
    metric : an instance of the class Metrics2d containing the gradient operators
    absorption_coeff : absorption coefficient of shape (size_ns * size_we)
    spectral_radiance : field of spectral radiation (size_ns * size_we)

    Returns
    -------
    LHS matrix
    """

    # Left Hand Side
    grad_x = metric.grad_x.toarray()
    grad_y = metric.grad_y.toarray()

    alpha = 1. / absorption_coeff
    beta = - 3. * absorption_coeff

    # div (alpha * grad(.)) operator
    alpha_mat = np.diag(alpha.ravel())
    a_grad_x = np.dot(alpha_mat, grad_x)
    a_grad_y = np.dot(alpha_mat, grad_y)
    term1 = np.dot(grad_x, a_grad_x) + np.dot(grad_y, a_grad_y)

    # (beta * .) operator
    term2 = np.diag(beta.ravel())

    lhs = term1 + term2

    # Right-Hand Side
    abs_gas = absorption_coeff.ravel()
    spec_rad = spectral_radiance.ravel()
    rhs = -12. * np.pi * abs_gas * spec_rad

    return lhs, rhs


def get_p1_eq_field_csr(metric, absorption_coeff, spectral_radiance):
    """
    Create the LHS matrix corresponding to P1 equation
    Book --> Radiative Heat Transfer (Modest, p. 479)
    Parameters
    ----------
    metric : an instance of the class Metrics containing the gradient operators
    absorption_coeff : absorption coefficient of shape (size_ns * size_we)
    spectral_radiance : field of spectral radiation (size_ns * size_we)

    Returns
    -------
    LHS matrix
    """

    # Left Hand Side
    grad_x = metric.grad_x_csr
    grad_y = metric.grad_y_csr

    alpha = 1. / absorption_coeff
    beta = - 3. * absorption_coeff

    # div (alpha * grad(.)) operator
    alpha_mat = scp.csr_matrix(np.diag(alpha.ravel()))
    a_grad_x = alpha_mat.dot(grad_x)
    a_grad_y = alpha_mat.dot(grad_y)
    term1 = grad_x.dot(a_grad_x) + grad_y.dot(a_grad_y)

    # (beta * .) operator
    term2 = scp.csr_matrix(np.diag(beta.ravel()))

    lhs = term1 + term2

    # Right-Hand Side
    abs_gas = absorption_coeff.ravel()
    spec_rad = spectral_radiance.ravel()
    rhs = -12. * np.pi * abs_gas * spec_rad

    return lhs, rhs


def analytical_temperature(abs_coeff, radius, temp_1, temp_2):
    """
    Analytical temperature profile for P1.
    Book --> Radiative Heat Transfer (Modest, p. 479)

    Parameters
    ----------
    abs_coeff : absorption coefficient
    radius : radial coordinates
    temp_1 : temperature of inner cylinder
    temp_2 : temperature of outer cylinder

    Returns
    -------
    Analytical temperature field
    """
    tau1 = abs_coeff * np.amin(radius)
    tau2 = abs_coeff * np.amax(radius)
    tau = abs_coeff * radius

    num = 1. + 3. / 2. * tau2 * np.log(tau2 / tau)
    denum = 1. + tau2 / tau1 + 3. / 2. * tau2 * np.log(tau2 / tau1)

    phi = ((temp_1 ** 4. - temp_2 ** 4.) * num / denum + temp_2 ** 4.) ** (
            1. / 4.)

    return phi


def dump_solution(metric, incident_radiation_sol, S_r_sum, S_r_div,
                  analytic_spectral_radiance, analytic_temperature, q_r_flux):
    """
    Dump generic postprocessing data

    Parameters
    ----------
    metric: metric object
    incident_radiation_sol: G field
    S_r_sum: radiative source term computed using method 1 (cf function)
    S_r_div: radiative source term computed using method 2 (cf function)
    analytic_spectral_radiance: Ib calculated from analytic_temperature
    analytic_temperature: initial temperature
    q_r_flux: radiative heat flux (x, y and normal)

    Returns
    -------
    None
    """
    out_h5 = NpArray2Xmf("radiation_solution.h5")
    out_h5.create_grid(metric.x_coor, metric.y_coor, np.zeros(metric.shp2d))

    # Inputs
    out_h5.add_field(analytic_spectral_radiance, "Ib_gas")
    out_h5.add_field(analytic_temperature, "T_init")

    # Output
    out_h5.add_field(incident_radiation_sol, "G")

    # Post-processing data
    q_r_x = q_r_flux[0].reshape(metric.shp2d)
    q_r_y = q_r_flux[1].reshape(metric.shp2d)
    q_r_n = q_r_flux[2].reshape(metric.shp2d)
    out_h5.add_field(S_r_sum, "Sr")
    out_h5.add_field(S_r_div.reshape(metric.shp2d), "S_r_div")
    out_h5.add_field(q_r_x, "Heat flux x")
    out_h5.add_field(q_r_y, "Heat flux y")
    out_h5.add_field(q_r_n, "Heat flux normal")
    out_h5.dump()


def dump_solution_test_case(metric, q_r_flux, params):
    """
    Dump Analytical-test-case-related data.

    Parameters
    ----------
    metric: metric object
    q_r_flux: heat flux
    params: dict of parameters

    Returns
    -------
    None
    """
    # Analytical test case
    out_h5 = NpArray2Xmf("radiation_solution_test_case.h5")
    out_h5.create_grid(metric.x_coor, metric.y_coor, np.zeros(metric.shp2d))
    adim_flux_x = adim_flux_test_case(q_r_flux[0].reshape(metric.shp2d),
                                      params)
    adim_flux_y = adim_flux_test_case(q_r_flux[1].reshape(metric.shp2d),
                                      params)
    adim_flux_n = adim_flux_test_case(q_r_flux[2].reshape(metric.shp2d),
                                      params)
    out_h5.add_field(adim_flux_x, "Adim heat flux x")
    out_h5.add_field(adim_flux_y, "Adim heat flux y")
    out_h5.add_field(adim_flux_n, "Adim heat flux normal")
    out_h5.dump()
    pass


def get_heat_flux_inner(x_coor, y_coor, q_n):
    rad = np.sqrt(x_coor ** 2. + y_coor ** 2.)
    r_min = 1.
    tol = 1e-7

    mask = np.where(rad < r_min + tol)
    idx = np.argmin(x_coor[mask])
    heat_flux_inner = q_n[mask]

    return float(-1. * heat_flux_inner[idx])


def get_heat_source_term(g_solution, absorption_coefficient,
                         spectral_radiance):
    """
    Computes radiative heat source term

    Parameters
    ----------
    g_solution : solution of incident radiation
    absorption_coefficient: field of absorption coefficient
    spectral_radiance: field of spectral radiance

    Returns
    -------
    heat source term
    """
    print("\t>> Compute radiative energy source term")
    radiative_source_term = absorption_coefficient * (
            4. * np.pi * spectral_radiance - g_solution)

    return radiative_source_term


def get_source_term_div(heat_fluxes, metric):
    """
    Compute the radiative in the conventional way by taking the divergence of
    the radiative heat flux vector

    Parameters
    ----------
    heat_fluxes: numpy arrays containing heat fluxes in x, y, and projected
    onto the boundaries
    metric: object containing information relative to metric (gradients,
    coordinates, etc.)

    Returns
    source_term: radiative source term
    -------

    """
    q_x = heat_fluxes[0]
    q_y = heat_fluxes[1]

    grad_x = metric.grad_x_csr
    grad_y = metric.grad_y_csr
    radiative_source_term_div = grad_x.dot(q_x) + grad_y.dot(q_y)

    return radiative_source_term_div


def get_heat_fluxes(absorption_coefficient, metric, g_solution):
    """
    Computes radiative heat fluxes

    Parameters
    ----------
    absorption_coefficient
    metric : Metrics object
    g_solution : solution of incident radiation

    Returns
    -------

    x and y heat fluxes
    """

    print("\n\t\t>> Compute heat fluxes")
    alpha = 1. / absorption_coefficient

    gamma = np.diag((-alpha.ravel() / 3.))
    gamma_csr = scp.csr_matrix(gamma)

    gamma_grad_x = gamma_csr.dot(metric.grad_x_csr)
    gamma_grad_y = gamma_csr.dot(metric.grad_y_csr)

    g_sol_csr = g_solution.ravel()

    heat_flux_x = gamma_grad_x.dot(g_sol_csr)
    heat_flux_y = gamma_grad_y.dot(g_sol_csr)

    heat_flux_n_bc = metric.n_x.dot(heat_flux_x) + metric.n_y.dot(
        heat_flux_y)

    return heat_flux_x, heat_flux_y, heat_flux_n_bc


def analytical_test_case(metric, params):
    """
    This functions returns fields corresponding to the 2 concentric
    cylinder with grey gas (Modest's book, 2nd edition, p. 477)

    Parameters
    ----------
    metric: metric object
    params: dict of parameters

    Returns
    -------
    analytic_temperature: temperature field
    analytic_spectral_radiance: spectral radance compute from temperature
    absorption_field: absorption coefficient field
    """

    t_west = params["boundaries"]["West"]["Values"]["Wall_temperature"]
    t_east = params["boundaries"]["East"]["Values"]["Wall_temperature"]

    radius = np.sqrt(metric.x_coor ** 2. + metric.y_coor ** 2.)
    print("\n\t--> Compute temperature field")

    analytic_temperature = analytical_temperature(
        params["field"]["absorption_coefficient"], radius, t_west, t_east)

    print("\n\t--> Compute spectral radiance")
    analytic_spectral_radiance = get_spectral_radiance(analytic_temperature)

    kappa_cst = params["field"]["absorption_coefficient"]
    absorption_field = np.ones_like(metric.x_coor) * kappa_cst

    return analytic_temperature, analytic_spectral_radiance, absorption_field


def temperature_slope(x_coor, y_coor, t_hot, t_cold, mesh_kind):
    """
    Compute initial temperature field as a temperature slope
    Parameters
    ----------
    x_coor:  x coordinates
    y_coor: y coordinates
    t_hot: higher value of temperature
    t_cold: lower value of temperature
    mesh_kind: mesh kind (either 'cyl' or 'rect'

    Returns
    -------
    temperature field (same shape that x_coor)
    """

    temperature = np.zeros_like(x_coor)

    if mesh_kind == "cyl":
        radius = np.sqrt(np.square(x_coor) + np.square(y_coor))
        temperature = (t_hot - t_cold) * (radius - 1) + t_cold
    elif mesh_kind == "rect":
        temperature = (t_hot - t_cold) * x_coor + t_cold

    return temperature


def get_spectral_radiance(temp):
    """
    Compute blackbody thermal radiation intensity
    Parameters
    ----------
    temp : temperature

    Returns
    -------
    blackbody thermal radiation intensity
    """
    sigma = 5.670373e-8
    return temp ** 4. / np.pi * sigma


def postproc_radiation(metric, incident_radiation_sol, absorption_field,
                       analytic_spectral_radiance):
    """
    Generic post proc for P1 solver

    Parameters
    ----------
    metric: object containing metrics (coord, gradients, etc.)
    incident_radiation_sol: G solution
    absorption_field: kappa field
    analytic_spectral_radiance: Ib for test case

    Returns
    -------
    Radiative heat fluxes and source terms
    """

    heat_fluxes = get_heat_fluxes(absorption_field, metric,
                                  incident_radiation_sol)

    # There are 2 ways of computing the radiative source term
    # Way 1 : Eq. (5) in :
    #
    #   "C. Paul, D. C. Haworth, and M. F. Modest. A simplified CFD model
    #    for spectral radiative heat transfer in high-pressure
    #    hydrocarbon-air combustion systems. Proc. Combust. Inst.,
    #    000 :1–8, 2018.
    radiative_source_term = get_heat_source_term(incident_radiation_sol,
                                                 absorption_field,
                                                 analytic_spectral_radiance)

    # Way 2 : Generic definition of the flux S_r = - div(q_r)
    #
    #   "M.F. Modest , D.C. Haworth , Radiative Heat Transfer in Turbulent
    #    Combustion Systems , Springer, 2016."
    radiative_source_term_div = get_source_term_div(heat_fluxes, metric)

    return heat_fluxes, radiative_source_term, radiative_source_term_div


def postproc_analytical_test_case(metric, params, heat_fluxes):
    """
    This functions returns the non dimensional normal heat flux of the inner
    cylinder in the analytices test case (2 concentric cylinders with grey gas
    (Modest's book, 2nd edition, p. 477))

    Returns
    -------
    Non-dimensional normal heat flux at the inner cylinder wall.
    """
    q_n = adim_flux_test_case(heat_fluxes[2].reshape(metric.shp2d), params)
    heat_flux_scalar = get_heat_flux_inner(metric.x_coor, metric.y_coor, q_n)

    return heat_flux_scalar


def adim_flux_test_case(dim_data, params):
    """
    Non-dimensional conversion of heat fluxes

    Parameters
    ----------
    dim_data: dimensional data
    params: parameter dictionary

    Returns
    -------
    non dimensional data
    """

    def adim_factor(temp1, temp2):
        """
        Factor use to non-dimensionalize results

        Parameters
        ----------
        temp1: temperature of cylinder one
        temp2: temperature of cylinder two

        Returns
        -------
        factor to adimensionalize data as in analytical test case
        """
        sigma = 5.670373e-8
        return 1. / (sigma * (temp1 ** 4. - temp2 ** 4.))

    t_west = params["boundaries"]["West"]["Values"]["Wall_temperature"]
    t_east = params["boundaries"]["East"]["Values"]["Wall_temperature"]
    factor = adim_factor(t_west, t_east)
    return dim_data * factor


def compute_robin_values_p1(absorption_coeff, t_wall, emissivity=1):
    """
    Compute values of the Marshak boundary condition in the Robin formalism:

                    a * (df / dn) + b * f + c = O

    Parameters
    ----------
    absorption_coeff: absorption coefficient field
    t_wall: temperature of the boundary wall
    emissivity: emissivity of the wall (1 if it is a black body)

    Returns
    -------
    a_value: pre-normal-derivative coefficient
    b_value: f-linear coefficient
    c_value: constant coefficient (does not depend on f)
    """

    a_value = (2. - emissivity) / emissivity * 2. / (
            3. * absorption_coeff)
    b_value = 1.
    c_value = - 4. * np.pi * get_spectral_radiance(t_wall)

    return a_value, b_value, c_value


def set_bc_p1(bc_params, absorption_coeff):
    """
    Apply Marshak boundary conditions for the P1 solver from physical params

    Parameters
    ----------
    bc_params: dictionary of boundary parameters
    absorption_coeff: absorption coefficient field

    Returns
    -------
    bc_params: updated dictionary of boundary parameters
    """

    for boundary_name, bc_dict in bc_params.items():

        if bc_dict["type"] == 'Robin':
            t_wall = bc_dict["Values"]["Wall_temperature"]

            if "Emissivity" in bc_dict["Values"].keys():
                emissivity = bc_dict["Values"]["Emissivity"]
                bc_val = compute_robin_values_p1(absorption_coeff, t_wall,
                                                 emissivity=emissivity)
            else:
                # Treat wall as black wall by default
                bc_val = compute_robin_values_p1(absorption_coeff, t_wall)

            bc_dict["bc_values"] = bc_val

    return bc_params


def main_p1_solver(params):
    """
    Radiation solver (P1 Approximation)

    Parameters :
    ------------
    params : dict of setup parameters

    """
    print("\n\tmain ---> Generate Grid")
    x_coor, y_coor = msh.get_mesh(params["mesh"])

    if "dilate_grid" in params.keys():
        if params["dilate_grid"]:
            x_coor, y_coor, _ = msh.dilate_center(x_coor, y_coor)

    print("\n\tmain ---> Define Boundaries")
    bnd = bndy.Boundary2d(params["boundaries"])

    print("\n\tmain ---> Compute Metrics")
    metric = Metrics2d(x_coor, y_coor, periodic_ns=bnd.periodic_ns,
                       periodic_we=bnd.periodic_we)

    fields = analytical_test_case(metric, params)
    analytic_temperature, analytic_spectral_radiance, absorption_field = fields

    print("\n\tmain --> Build LHS and RHS")
    lhs_csr, rhs_csr = get_p1_eq_field_csr(metric, absorption_field,
                                           analytic_spectral_radiance)

    print("\n\tmain ---> Apply BC")
    # We take the mean value of the absorption_field so far (ok for grey gas)
    params["boundaries"] = set_bc_p1(params["boundaries"],
                                     np.mean(absorption_field))

    lhs_csr_bc, rhs_csr_bc = bndy.apply_bc(lhs_csr, rhs_csr, metric,
                                           params["boundaries"])

    print("\n\tmain ---> Solving...")
    solution_1d, info = linalg.bicgstab(lhs_csr_bc, rhs_csr_bc, atol='legacy')
    incident_radiation_sol = solution_1d.reshape(metric.shp2d)

    print("\n\tmain ---> Postproc...")
    q_r_flux, s_r_sum, s_r_div = postproc_radiation(metric,
                                                    incident_radiation_sol,
                                                    absorption_field,
                                                    analytic_spectral_radiance)

    heat_flux_scalar = postproc_analytical_test_case(metric, params, q_r_flux)

    print("\n\tmain ---> Write Solution")
    dump_solution(metric, incident_radiation_sol, s_r_sum, s_r_div,
                  analytic_spectral_radiance, analytic_temperature, q_r_flux)

    dump_solution_test_case(metric, q_r_flux, params)

    results = {}
    results['solution_1d'] = solution_1d
    results['incident_radiation_sol'] = incident_radiation_sol
    results['q_r_flux'] = q_r_flux
    results['s_r_sum'] = s_r_sum
    results['s_r_div'] = s_r_div
    results['heat_flux_scalar'] = heat_flux_scalar

    print("\n\tmain --> Done! =)")

    return heat_flux_scalar, results


if __name__ == "__main__":
    T_west = 500.
    T_east = 1500.

    # Mesh definition
    MESH_RECT = {"kind": 'rect',
                 "x_max": 1.,
                 "y_max": 1.,
                 "x_res": 5.,
                 "y_res": 6.,
                 }

    MESH_CYL = {"kind": 'cyl',
                "r_min": 1.,
                "r_max": 2.,
                "theta_min": 0.,
                "theta_max": 2. * np.pi,
                "n_pts_rad": 15
                }

    # Boundary conditions
    # Default
    BOUNDARY = {"North": {"type": "Periodic", 'Values': {}},
                "West": {"type": "Periodic", 'Values': {}},
                "South": {"type": "Periodic", 'Values': {}},
                "East": {"type": "Periodic", 'Values': {}}}

    BOUNDARY["West"]["type"] = "Robin"
    BOUNDARY["West"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 500.}

    BOUNDARY["East"]["type"] = "Robin"
    BOUNDARY["East"]["Values"] = {"Emissivity": 1.,
                                  "Wall_temperature": 1500.}

    # Input fields: Temperature, pressure, etc.
    FIELD = {"absorption_coefficient": 1.,
             "temperature": [T_west, T_east]}

    # Create parameter dictionary
    PARAMS = {}
    PARAMS["boundaries"] = BOUNDARY
    PARAMS["field"] = FIELD

    # Setup simulation with rectangular mesh
    # PARAMS["mesh"] = MESH_RECT

    # Setup simulation with cylindrical mesh
    PARAMS["mesh"] = MESH_CYL

    # emissivity = PARAMS["boundaries"]["Values"]["Emissivity"]
    # t_wall = PARAMS["boundaries"][["Values"]["Wall_temperature"]

    q_inner = main_p1_solver(PARAMS)



""" Module concerning some 3D vector manipulations in numpy

OST :Mercy in Darkness, by Two Steps From Hell

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

SMALL = 1e-12


def renormalize(np_ar_vect):
    """ renormalize a numpy array of vectors,
    considering the last axis """
    result = (np_ar_vect
              / np.expand_dims(np.linalg.norm(np_ar_vect, axis=-1),
                               axis=-1))
    return result


def angle_btw_vects(np_ar_vect1, np_ar_vect2, convert_to_degree=False):
    """ compute the angle in deg btw two UNIT vectors """
    sdot = np.sum(renormalize(np_ar_vect1)
                  * renormalize(np_ar_vect2), axis=-1)

    # Clipping is needed when pscal.max() > 1 : 1.000000000002
    sdot = np.clip(sdot, -1.0, 1.0)
    angles = np.arcsin(sdot)
    if convert_to_degree:
        angles = np.rad2deg(angles)
    return angles


def yz_to_theta(np_ar_vect):
    """ return theta , a radians measure of ange in the yz plane,
    - range -pi/pi
    - 0 on the y+ axis (z=0, y>0)
    spanning -pi to pi
                   0pi=0deg
                  Y
                  ^
                  |
                  |
-0.5pi=-90deg     o------>Z   0.5pi=90deg

    """
    # return np.arctan2(np_ar_vect[:, 2],
    #                   np_ar_vect[:, 1])
    return np.arctan2(np.take(np_ar_vect, 2, axis=-1),
                      np.take(np_ar_vect, 1, axis=-1))


def rtheta2yz(rrr, theta):
    """ return yz fror rtheta ,
    theta in  radians measure of ange in the yz plane,
    - range -pi/pi
    - 0 on the y+ axis (z=0, y>0)
    spanning -pi to pi
                   0pi=0deg
                  Y
                  ^
                  |
                  |
-0.5pi=-90deg     o------>Z   0.5pi=90deg

    """
    yyy = rrr * np.cos(theta)
    zzz = rrr * np.sin(theta)
    return yyy, zzz


def rotate_vect_around_x(np_ar_vect, angle_deg):
    """ rotate vector around axis x in degree """

    arr_r = np.hypot(np.take(np_ar_vect, 1, axis=-1),
                     np.take(np_ar_vect, 2, axis=-1))
    arr_theta = (yz_to_theta(np_ar_vect)
                 + np.deg2rad(angle_deg))

    result = np.stack((np.take(np_ar_vect, 0, axis=-1),
                       arr_r * np.cos(arr_theta),
                       arr_r * np.sin(arr_theta)),
                      axis=-1)

    return result


def mask_cloud(np_ar_xyz,
               axis,
               support):
    """ mask a cloud of n 3D points in in xyz
    axis among x,y,z,theta,r
    support a 2 value tuple : (0,3), (-12,float(inf))

    x and (0,3) reads as  0 <= x < 3 ( lower bound inclusive)
    z and (-12,float(inf)) reads as  -12 <= z
    theta in degree, cyl. coordinate around x axis
    - range -180,180
    - 0 on the y+ axis (z=0, y>0)
    """

    def true_between(np_arr, low, high):
        """ return a boolean array of the same shape,
        true if value are btw low and high"""
        return (np_arr >= low) * (np_arr < high)

    possible_axes = ["x", "y", "z", "theta", "r"]
    if axis not in possible_axes:
        raise IOError("axis " + axis + "not among possible axes "
                      + ";".join(possible_axes))

    if axis == "x":
        mask = true_between(
            np.take(np_ar_xyz, 0, axis=-1),
            support[0],
            support[1])
    if axis == "y":
        mask = true_between(
            np.take(np_ar_xyz, 1, axis=-1),
            support[0],
            support[1])
    if axis == "z":
        mask = true_between(
            np.take(np_ar_xyz, 2, axis=-1),
            support[0],
            support[1])
    if axis == "r":
        mask = true_between(
            np.hypot(np.take(np_ar_xyz, 1, axis=-1),
                     np.take(np_ar_xyz, 2, axis=-1)),
            support[0],
            support[1])
    if axis == "theta":
        mask = true_between(
            np.rad2deg(yz_to_theta(np_ar_xyz)),
            support[0],
            support[1])
    return mask



""" small timer function """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time


def timing(function):
    """ lazy method to time my function """

    def wrap(*args, **kw):
        """ wrap the function to time it"""
        time1 = time.time()
        ret = function(*args, **kw)
        time2 = time.time()
        print("----- %s took %0.3f s" % (function.func_name, (time2 - time1)))
        return ret

    return wrap



""" module to create an ensight compatible file
to visualize your  data"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

import h5py
import numpy as np
from lxml import etree

NSMAP = {"xi": "http://www.w3.org/2001/XInclude"}

class NpArray2Xmf():
    """ main class for data output in XDMF format """

    def __init__(self, filename, domain_name=None, mesh_name=None, time=None):
        """ class startup"""
        extension = os.path.splitext(filename)
        if extension[-1] == ".xmf":
            self.filename = extension[0] + ".h5"
        elif extension[-1] == ".h5":
            self.filename = filename
        else:
            raise RuntimeError("Only extensions .xmf or .h5 are allowed")

        self.geotype = None
        self.mesh = {}
        self.data = {}
        self.shape = None
        self.mesh["domain"] = domain_name
        self.mesh["mesh"] = mesh_name
        self.mesh["time"] = time
        if time is None:
            self.mesh["time"] = 0.0
        self.mesh["time"] = "%14.8e" %self.mesh["time"]

        if self.mesh["mesh"] is None:
            self.mesh["mesh"] = "Mesh"

        if self.mesh["domain"] is None:
            self.mesh["domain"] = "Domain"


    def create_grid(self, nparray_x, nparray_y, nparray_z):
        """ create the grid according to numpy arrays x, y ,z
        if arrays are 1D, switch to cloud point
        if arrays are 2D, switch to quad connectivity
        if arrays are 3D, switch to hexaedrons connectivity"""
        self.mesh["x"] = np.ravel(nparray_x)
        self.mesh["y"] = np.ravel(nparray_y)
        self.mesh["z"] = np.ravel(nparray_z)
        self.shape = list(nparray_x.shape)
        dim = len(self.shape)
        if dim == 1:
            self.geotype = "cloud"
        if dim == 2:
            self.geotype = "quads"
        if dim == 3:
            self.geotype = "hexas"

        if self.geotype is None:
            raise RuntimeError("Unexpected shape of nparray :"
                               + " ".join(self.shape))

    def add_field(self, nparray_field, variable_name):
        """ add a field, assuming same shape as nparray of coordiantes """
        self.data[variable_name] = nparray_field

    def _type(self, var):
        """ retrun the xmf type according to nparray"""
        var_shape = list(self.data[var].shape)

        dtype = self.data[var].dtype
        numbertype = None
        if dtype == "float64":
            numbertype = "Float"
        if dtype == "S4":
            numbertype = "Char"
        if numbertype is None:
            raise RuntimeError("Array of type " + dtype
                               + "(" + var + ")"
                               + "not recognized")

        attributetype = None
        if var_shape == self.shape:
            attributetype = "Scalar"
        if var_shape[:-1] == self.shape:
            if var_shape[-1] == 3:
                attributetype = "Vector"
        if attributetype is None:
            raise RuntimeError("Var " + var
                               + " of shape " + str(var_shape)
                               + " not consistent  with grid of shape "
                               + str(self.shape) +
                               "\n (neither scalar nor 3D vector...)")

        shape_str = " ".join(str(dim) for dim in self.data[var].shape)
        return (numbertype,
                attributetype,
                shape_str)

    def xmf_dump(self):
        """ create XDMF descriptor file """
        if self.geotype == "cloud":
            topology = "PolyVertex"
        if self.geotype == "quads":
            topology = "2DSMesh"
        if self.geotype == "hexas":
            topology = "3DSMesh"

        dims = " ".join(str(dim) for dim in self.shape)

        xmf_tree = dict()
        xmf_tree['root'] = etree.Element("Xdmf", Version="2.0", nsmap=NSMAP)

        xmf_tree['dom'] = etree.SubElement(xmf_tree['root'], "Domain",
                                           Name=self.mesh['domain'])

        xmf_tree['grd'] = etree.SubElement(xmf_tree['dom'], "Grid",
                                           Name=self.mesh['mesh'],
                                           Type='Uniform')

        etree.SubElement(xmf_tree['grd'], "Time",
                         Type='Single',
                         Value=self.mesh['time'])

        etree.SubElement(xmf_tree['grd'], "Topology",
                         Name="Topo",
                         TopologyType=topology,
                         NumberOfElements=dims)

        xmf_tree['geo'] = etree.SubElement(xmf_tree['grd'], "Geometry",
                                           GeometryType="X_Y_Z")

        for var in ['x', 'y', 'z']:
            field = etree.SubElement(xmf_tree['geo'], "DataItem",
                                     Dimensions=dims,
                                     Format="HDF",
                                     NumberType="Float",
                                     Precision="8")
            text = "%s:/mesh/%s" %(os.path.basename(self.filename), var)
            field.text = "\n%s%s\n%s" %(11*" ", text, 8*" ")

        for var in self.data:
            numbertype, attributetype, dims = self._type(var)

            attr = etree.SubElement(xmf_tree['grd'], "Attribute",
                                    Name=var,
                                    Center="Node",
                                    AttributeType=attributetype)

            field = etree.SubElement(attr, "DataItem",
                                     Dimensions=dims,
                                     Format="HDF",
                                     NumberType=numbertype,
                                     Precision="8")
            text = "%s:/variables/%s" %(os.path.basename(self.filename), var)
            field.text = "\n%s%s\n%s" %(11*" ", text, 8*" ")

        xmf_file = self.filename.replace(".h5", ".xmf")
        xmf_ct = etree.tostring(xmf_tree['root'],
                                pretty_print=True,
                                xml_declaration=True,
                                doctype='<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>')
        xmf_ct = xmf_ct.decode().replace("encoding=\'ASCII\'", "")
        with open(xmf_file, "w") as fout:
            fout.write(xmf_ct)

    def dump(self):
        """ dump the final file """
        fout = h5py.File(self.filename, "w")

        mesh_gp = fout.create_group("mesh")
        for coord in ["x", "y", "z"]:
            mesh_gp.create_dataset(coord, data=self.mesh[coord])

        var_gp = fout.create_group("variables")
        for var in self.data:
            var_gp.create_dataset(var, data=self.data[var])
        fout.close()
        self.xmf_dump()


def create_time_collection_xmf(collection_filenames, xmf_filename):
    """ Creates xmf file holding time collection of xmf files

    Parameters :
    ==========
    collection_filenames: a list of single time xmf filenames to collect
    xmf_filename : the name of the output file

    Returns:
    ======
    None
    """

    root = etree.Element("Xdmf", Version="2.0", nsmap=NSMAP)
    dom = etree.SubElement(root, "Domain")

    grid = etree.SubElement(dom, "Grid",
                            Name=os.path.split(xmf_filename)[-1],
                            GridType="Collection",
                            CollectionType="Temporal")

    for filename in collection_filenames:
        etree.SubElement(grid, "XI_INCLUDE", href=filename,
                         xpointer='xpointer(//Xdmf/Domain/Grid)')

    xmf_ct = etree.tostring(root,
                            pretty_print=True,
                            xml_declaration=True,
                            doctype='<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>')
    xmf_ct = xmf_ct.decode()
    with open(xmf_filename, "w") as fout:
        xmf_ct = xmf_ct.replace("encoding=\'ASCII\'", "")
        xmf_ct = xmf_ct.replace("XI_INCLUDE", "xi:include")
        fout.write(xmf_ct)


if __name__ == '__main__':

    DIM_X = 41
    DIM_Y = 21
    DIM_Z = 11

    SIZE_X = 4.
    SIZE_Y = 2.
    SIZE_Z = 1.

    # 1D
    TEST_X = np.linspace(0, SIZE_X, DIM_X)
    TEST_Y = np.linspace(0, SIZE_Y, DIM_X)
    TEST_Z = np.linspace(0, SIZE_Z, DIM_X)
    TEST_U = (np.sin(TEST_X / SIZE_X * 1 * np.pi)
              * np.sin(TEST_Y / SIZE_Y * 1 * np.pi)
              * np.sin(TEST_Z / SIZE_Z * 1 * np.pi))

    TEST_F = NpArray2Xmf("./test1D.h5")
    TEST_F.create_grid(TEST_X, TEST_Y, TEST_Z)
    TEST_F.add_field(TEST_U, "foobar")

    TEST_V = np.stack((TEST_U,
                       TEST_U,
                       TEST_U),
                      axis=1)
    TEST_F.add_field(TEST_V, "foobar_vect")

    TEST_F.dump()

    # 2D
    TEST_X = np.tile(np.linspace(0., SIZE_X, DIM_X), (DIM_Y, 1))
    TEST_Y = np.tile(np.linspace(0., SIZE_Y, DIM_Y), (DIM_X, 1)).transpose()
    TEST_Z = np.ones((DIM_Y, DIM_X))
    TEST_U = (np.sin(TEST_X / SIZE_X * 1 * np.pi)
              * np.sin(TEST_Y / SIZE_Y * 1 * np.pi)
              * np.sin(TEST_Z * 0.5 * np.pi))

    TEST_F = NpArray2Xmf("./test2D.h5")
    TEST_F.create_grid(TEST_X, TEST_Y, TEST_Z)
    TEST_F.add_field(TEST_U, "foobar")
    TEST_F.dump()

    TEST_X = TEST_X[:, :, None].repeat(DIM_Z, 2)
    TEST_Y = TEST_Y[:, :, None].repeat(DIM_Z, 2)
    TEST_Z = np.tile(np.linspace(0., SIZE_Z, DIM_Z), (DIM_X, 1)).transpose()
    TEST_Z = TEST_Z[:, :, None].repeat(DIM_Y, 2).transpose((2, 1, 0))
    TEST_U = (np.sin(TEST_X / SIZE_X * 1 * np.pi)
              * np.sin(TEST_Y / SIZE_Y * 1 * np.pi)
              * np.sin(TEST_Z / SIZE_Z * 1 * np.pi))

    TEST_F = NpArray2Xmf("./test3D.h5")
    TEST_F.create_grid(TEST_X, TEST_Y, TEST_Z)
    TEST_F.add_field(TEST_U, "foobar")
    TEST_F.dump()



"""
This module contains function to create and modify meshes
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np


def gen_cart_grid_2d(gridrange, gridpoints):
    """ Generate cartesian grid.

    Parameters :
    ------------
    gridrange : tuple of floats, dimensions of the grid
    gridpoints : tuple of ints (n,m), sampling on the grid

    Returns :
    ---------

    x_coor, y_coor : numpy arrays (n,m) with coordinates
    """
    x_vec = np.linspace(0., gridrange[0], int(gridpoints[0]), endpoint=True)
    y_vec = np.linspace(0., gridrange[1], int(gridpoints[1]), endpoint=True)
    x_coor, y_coor = np.meshgrid(x_vec, y_vec)

    return x_coor, y_coor


def gen_cyl_grid_2d(r_min, r_max, theta_min, theta_max, r_points, theta_points):
    """
    Generate a cylindrical grid center on x = 0 and y = 0

    Parameters
    ----------
    r_min : inner radius
    r_max : outer radius
    theta_min : lower angle [0, 2 * pi]
    theta_max : upper angle [0, 2 * pi]
    r_points : number of points in the radial direction
    theta_points : number of points in the tangential direction

    Returns
    -------
    x_coor : x coordinates of the mesh
    y_coor : y coordinates of the mesh
    """
    r_vec = np.linspace(r_min, r_max, r_points, endpoint=True)
    theta_vec = np.linspace(theta_min, theta_max, theta_points, endpoint=True)
    r_coor, theta_coor = np.meshgrid(r_vec, theta_vec)

    x_coor = r_coor * np.cos(theta_coor)
    y_coor = r_coor * np.sin(theta_coor)

    return x_coor, y_coor


def dilate_center(x_coor, y_coor, perturbation=0.1):
    """ perturb cartesian mesh dilatation in the center

    Parameters :
    -----------
    x_coor : numpy array (n,m) , x_coordinates
    y_coor : numpy array (n,m) , y_coordinates
    perturbation : float, amplitude of the perturbation perturbation
                   with respect to the grid size

    Returns :
    ----------
    x_coor : numpy array (n,m) , x_coordinates shifted
    y_coor : numpy array (n,m) , y_coordinates shifted

     """
    x_width = x_coor.max() - x_coor.min()
    y_width = y_coor.max() - y_coor.min()
    x_center = x_coor.min() + 0.5 * x_width
    y_center = y_coor.min() + 0.5 * y_width

    adim_rad = np.sqrt(((x_coor - x_center) / (0.5 * x_width)) ** 2
                       + ((y_coor - y_center) / (0.5 * y_width)) ** 2)

    rad_dir_x = (x_coor - x_center) / np.maximum(adim_rad, 1.e-16)
    rad_dir_y = (y_coor - y_center) / np.maximum(adim_rad, 1.e-16)
    dilate = np.maximum((1. - adim_rad), 0)
    dilate = 3 * dilate * dilate * (1 - dilate)

    shift_x = perturbation * x_width * dilate * rad_dir_x
    shift_y = perturbation * y_width * dilate * rad_dir_y

    return x_coor + shift_x, y_coor + shift_y, dilate


def get_mesh(params_mesh):
    """
    Call specific meshing functions from mesh parameters dict

    Parameters
    ----------
    params_mesh: dictionary containing mesh parameters

    Returns
    -------
    x_coor: x coordinates of the mesh
    y_coor: y coordinates of the mesh
    """

    if params_mesh["kind"] == "rect":
        x_max = params_mesh["x_max"]
        y_max = params_mesh["y_max"]
        x_res = params_mesh["x_res"]
        y_res = params_mesh["y_res"]
        x_coor, y_coor = gen_cart_grid_2d((x_max, y_max), (x_res, y_res))

    elif params_mesh["kind"] == "cyl":
        r_min = params_mesh["r_min"]
        r_max = params_mesh["r_max"]
        theta_min = params_mesh["theta_min"]
        theta_max = params_mesh["theta_max"]
        angle = theta_max - theta_min
        n_r_pts = params_mesh["n_pts_rad"]

        try:
            n_tgt_pts = params_mesh["n_pts_tgt"]
        except KeyError:
            n_tgt_pts = int(angle * n_r_pts + 1)

        delta = angle / (int(n_tgt_pts) - 1)
        theta_max -= delta

        x_coor, y_coor = gen_cyl_grid_2d(r_min, r_max, theta_min, theta_max,
                                         n_r_pts, int(angle * n_r_pts + 1))
    else:
        raise NotImplementedError(
            'The only mesh kinds implemented are: "rect" and "cyl"')

    return x_coor, y_coor



""" module to compute the directed projection
    of a point to a surface along a direction
         ___
            \
             \
        x---->\
               \
               
OST : Seven nation Army (Westworld), R. Djawadi

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from scipy import spatial
import numpy as np

from arnica.utils.axishell import AxiShell
from arnica.utils.timer_decorator import timing
from arnica.utils.nparray2xmf import NpArray2Xmf
from arnica.utils.vector_actions import renormalize

BIG = 1e6
# @timing


def projection_kdtree(
        points,
        directions,
        point_surface,
        normal_surface,
        neighbors=1000,
        tol=1000.0):
    """
    project the n points folowing the direction on the m suface
    points - "p"  nparray of shape (n,3),
    directions - nparray of shape (n,3),
    point_surface - "s" nparray of shape (m,3),
    normal_surface - nparray of shape (m,3),
    neigbors : nb of neigbors to take into account
    tol : maximum distance beyond cyl dist witl bi set to BIG

    return :
    points - "t"  nparray of shape (n,3), moved on the surface
    indexes - neigborhood of the points (n,k)
    cyl_dist - cylindrical distance of p with each neighbor (n,k)

           < shp_dist >
    ______s___________t__________________
          |' .                       A
          |     '  .< cyl_dist >     .
          v                          .< surface_dist>
                 4                   .
                /                    .
               /                     .
              p                      V

          align : alignment (pscal of two unit vectors, in [-1,1])

    1 - compute a kdtree, and reduce the surface
        points to a neighborhood indexes  [n,k]
    iterate over all points [n]:
    2 - compute the distance surface_dist to the surface, for each neighbor [k]
    3 - compute the alignment align, scal. product of normal and direction [k]
    4 - compute all points all_pts moved on the surface [k,3]
    5 - compute shperical distance  for each point to its moved point [k]
    6 - get the closest point idx closest_pt
    7 - recompute the cylindrical distance cyl_dist for all the neigborhood
    """
    neighbors = min(point_surface.shape[0], neighbors)
    print("KDTree with  "
          + "\n-   Surface points | "
          + str(point_surface.shape[0])
          + "\n- Projected points | "
          + str(points.shape[0])
          + "\n-        Neighbors | "
          + str(neighbors)
          )

    kdtree = spatial.cKDTree(point_surface)

    # first projection, one neigbor
    _, indexes = kdtree.query(points,
                              k=1)
    moved_pts = []
    for point, direction, i in zip(points, directions, indexes):
        pt_surf = point_surface[i, :]
        nml_surf = normal_surface[i, :]
        surface_dist = np.sum(np.multiply((pt_surf - point),
                                          nml_surf))
        align = np.sum(np.multiply(nml_surf, direction))
        mv_pt = point[:] + (surface_dist / align) * direction[:]
        moved_pts.append(mv_pt)
    moved_pts = np.asarray(moved_pts)

    # second projection, all neigbors
    _, indexes = kdtree.query(moved_pts,
                              k=neighbors)

    cyl_dists = []
    moved_pts = []
    for point, direction, indx in zip(points,
                                      directions,
                                      indexes):
        part_pts_surf = point_surface[indx, :]
        part_nml_surf = normal_surface[indx, :]

        surface_dist = np.sum(np.multiply((part_pts_surf - point),
                                          part_nml_surf),
                              axis=1)

        # scalar product of normals
        align = np.sum(np.multiply(part_nml_surf,
                                   direction),
                       axis=1)

        # all points projected
        all_pts = (point[np.newaxis, :]
                   + (surface_dist / align)[:, np.newaxis]
                   * direction[np.newaxis, :])

        sphere_dists = np.linalg.norm(part_pts_surf - all_pts, axis=1)

        closest_pt = all_pts[np.argmin(sphere_dists, axis=0), :]

        axis_dist = np.sum(np.multiply((part_pts_surf - closest_pt),
                                       direction),
                           axis=1)

        tmp_pts_surf = part_pts_surf - \
            axis_dist[:, np.newaxis]*direction[np.newaxis, :]
        cyl_dist = np.linalg.norm(tmp_pts_surf - closest_pt, axis=1)

        cyl_dist = np.where(align < 0.0, cyl_dist, BIG*np.ones(cyl_dist.shape))
        cyl_dist = np.where(sphere_dists < tol, cyl_dist,
                            BIG*np.ones(cyl_dist.shape))

        moved_pts.append(closest_pt)
        cyl_dists.append(cyl_dist)

    moved_pts = np.asarray(moved_pts)
    cyl_dists = np.asarray(cyl_dists)

    return (moved_pts,
            indexes,
            cyl_dists)



""" module loading avbp h5py into numpy arrays,
limited to point cloud (connectivity sucks, mark my word, really) """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import h5py
import numpy as np

from .vector_actions import renormalize


# TODO : premature optimization issue


class AVBPAsPointCloud(object):
    """ class handling mesh and solutions as point cloud
    *no connectivity asked* """

    def __init__(self, meshfile):
        """ startup class"""
        self.mesh = {}
        self.solavg = {}
        self.origins = {}
        self.origins["meshfile"] = meshfile
        self.origins["solavgfile"] = None

    # @timing
    def load_mesh_bulk(self):
        """ load the bulk of the mesh, withound the boundaries """

        with h5py.File(self.origins["meshfile"], "r") as fin:
            self.mesh["bulk"] = {}
            self.mesh["bulk"]["xyz"] = np.stack(
                (
                    fin["/Coordinates/x"][()],
                    fin["/Coordinates/y"][()],
                    fin["/Coordinates/z"][()],
                ),
                axis=1,
            )

    # @timing
    def load_mesh_bnd(self, patchlist=None):
        """ load only the boundaries. Load all patches,
        unless a subset of patch is provided with opt keyword patchlist """

        with h5py.File(self.origins["meshfile"], "r") as fin:
            self.mesh["Patches"] = {}

            patches_readable = fin["/Boundary/PatchLabels"][()]
            bnode_normal = fin["/Boundary/bnode->normal"][()]
            bnode_lidx = fin["/Boundary/bnode_lidx"][()]
            bnode_gnode = fin["/Boundary/bnode->node"][()]

            patches_readable = [pname.decode('UTF-8').strip() for pname in
                                patches_readable if pname is not None]

            if patchlist is None:
                patchlist = patches_readable
            else:
                for patch in patchlist:
                    if patch not in patches_readable:
                        raise IOError(
                            "No patch "
                            + patch
                            + " among :\n"
                            + "\n".join(patches_readable)
                        )

            for patchlabel in patchlist:
                self.mesh["Patches"][patchlabel] = {}

            for i, patchlabel in enumerate(patches_readable):
                if patchlabel in patchlist:
                    self.mesh["Patches"][patchlabel]["xyz"] = np.stack(
                        (
                            fin["/Patch/" + str(i + 1) + "/Coordinates/x"][()],
                            fin["/Patch/" + str(i + 1) + "/Coordinates/y"][()],
                            fin["/Patch/" + str(i + 1) + "/Coordinates/z"][()],
                        ),
                        axis=1,
                    )

                    shape = self.mesh["Patches"][patchlabel]["xyz"].shape
                    start = 0
                    end = 3 * bnode_lidx[i]
                    if i >= 1:
                        start = 3 * bnode_lidx[i - 1]
                    part_bnode_normal = bnode_normal[start:end].reshape(shape)
                    self.mesh["Patches"][patchlabel]["surf"] = np.linalg.norm(
                        part_bnode_normal, axis=-1
                    )
                    self.mesh["Patches"][patchlabel]["normal"] = renormalize(
                        part_bnode_normal
                    )

                    self.mesh["Patches"][patchlabel]["gnodes"] = (bnode_gnode[
                                                                  int(start / 3)
                                                                  : int(
                                                                      end / 3)])

    # @timing
    def load_avgsol(self, solavgfile):
        """ load a solution
        AVBP avg for the moment"""
        self.origins["solavgfile"] = solavgfile
        with h5py.File(self.origins["solavgfile"], "r") as fin:
            for var in ["T", "rho", "u", "v", "w"]:
                self.solavg[var] = fin["/Average/" + var][()]

    def get_skinpts(self, listpatch):
        """ return a dict of numpy array [x, y ,z]
        coordinates of a subset of patches """

        skin = np.concatenate(
            [self.mesh["Patches"][patchlabel]["xyz"] for patchlabel in
             listpatch])
        return skin



from __future__ import absolute_import, division
import math
import numpy as np

PRANDTL = 0.71
PRANDTL_TURB = 0.86
VKARMAN = 0.41
VAN_DRIEST = 5.5
BETA_KADER = ((3.85 * PRANDTL**(1.0/3.0) - 1.3)**2
              + 2.12 * math.log(PRANDTL))
K_KADER = ((BETA_KADER
            - PRANDTL_TURB * VAN_DRIEST
            + PRANDTL_TURB / VKARMAN - 2.12)
           - (PRANDTL_TURB / VKARMAN - 2.12)
           * (200.0*math.log(200.0) - 100*math.log(100.0))
           / (200.0 - 100.0))


def h_kader(t_wall,
            rho_wall,
            y_wall,
            u_2,
            t_2,
            temp_adiab):
    """ compute h at the wall as in kader
    names taken equalt to  loglaw_cwm.f90 AVBP"""
    _, heat_cp, mu_wall = lambda_cp_visco_fluid(t_2)
    # mu_wall : Kg / m.s
    # 1.0 Compute y_plus and u_tau
    # first estimation of y_plus
  
    yplus_guess = np.maximum(
        0.157 * (u_2 * y_wall * rho_wall / mu_wall)**(7.0/8.0),
        1.0e-3)
    #           [m/s *   m    *   Kg/m3  *s.m/ kg] = [-]
    # Linear part
    u_tau_lin = np.sqrt(u_2 / y_wall * mu_wall / rho_wall)
    #              sqrt[m/s /   m    * kg / m.s  *   m3/kg] =
    #              sqrt[m/s               /   s  *   m    ] =
    #              sqrt[m2/s2] = m/s

    #Log part
    def _get_u_tau_log():
        """ evaluate u tau in log region """
        u_tau_log = 0
        epsilon = 1.0
        yplus_log = yplus_guess

        def _loop_u_tau(u_tau_log, yplus_log):
            """ recursive loop to evaluate u_tau """
            utau_old = u_tau_log
            u_tau_log = u_2 / (1 / VKARMAN * np.log(yplus_log)
                               + VAN_DRIEST)
            yplus_log = u_tau_log * y_wall * rho_wall / mu_wall
            epsilon = np.abs(u_tau_log - utau_old)
            return u_tau_log, yplus_log, epsilon

        while np.min(epsilon) >= 1.e-5:
            (u_tau_log,
             yplus_log,
             epsilon) = _loop_u_tau(u_tau_log,
                                    yplus_log)

        return u_tau_log


    u_tau_log = _get_u_tau_log()
    # fusion of the two
    u_tau_cwm = np.where(yplus_guess < 11.25,
                         u_tau_lin,
                         u_tau_log)
    y_plus = u_tau_cwm * y_wall * rho_wall / mu_wall

    # 2.0 Compute T_tau
    kader_g = ((0.01 * (PRANDTL * y_plus)**4)
               / (1.0 + 5.0 * PRANDTL**3 * y_plus))
    t_tau_cwm = (
        (t_wall - t_2)
        / ((PRANDTL * y_plus * np.exp(-kader_g))
           + (PRANDTL_TURB*u_2/u_tau_cwm + K_KADER) * np.exp(-1.0/kader_g)))

    # unused tau_wall = rho_wall * u_tau_cwm * u_tau_cwm
    q_wall = rho_wall * heat_cp * u_tau_cwm * t_tau_cwm
    
    #print "qwall" ,t_wall.mean(),  (t_wall - t_2).min() , (t_wall - t_2).max()
    # resize h wall depending on the T adiab found -and not t_2-
    h_wall = -q_wall / (temp_adiab - t_wall)

    #raise ValueError("for some reason, the h coef is negative, i.e. your t_guess is too high")
    #h_wall = np.clip(h_wall, 1, 100000)

    return h_wall


def lambda_cp_visco_fluid(t_guess):
    """ compute Fluid properties lambda , cp, visco """
    def _visco_sutherland(t_guess):
        """ compute visocity as in sutherland"""
        mu_wall = (1.716e-5
                   * (t_guess / 273.15)**1.5
                   * (273.15 + 110.4) / (t_guess + 110.4))
        return mu_wall

    def _cp_correl(t_guess):
        """ compute cp of fluid"""
        # ADD CLIPPING
        #t_clip = np.max(t_guess, 300.0)
        #t_clip = np.min(t_clip, 2500.0)
        t_clip = t_guess

        heat_cp = (513.46
                   + 1.6837 * t_clip
                   - 0.0019275 * t_clip**2
                   + 1.2773e-6 * t_clip**3
                   - 4.2773e-10 * t_clip**4
                   + 5.6735e-14 * t_clip**5)
        return heat_cp
    heat_cp = _cp_correl(t_guess)
    mu_wall = _visco_sutherland(t_guess)
    lam = heat_cp / PRANDTL * mu_wall

    return lam, heat_cp, mu_wall



"""
Contains class and functions necessary to handle boundary conditions
"""

import numpy as np
import scipy.sparse as spsp
from arnica.solvers_2d.core_fd import set_row_csr


class Boundary2d:
    """Class containing Boundary Condition information"""

    def __init__(self, boundary_param):
        """
        Constructor of boundary class
        """
        self.boundary_params = boundary_param

        print('\n\t\t Boundary2D:\t _set_periodicity()')
        self._set_periodicity()

    def _set_periodicity(self):
        """
        Set the periodicity to North/South if periodic and check for the
        consistency of the definition of the BC
        """
        bc_n = self.boundary_params["North"]["type"]
        bc_w = self.boundary_params["West"]["type"]
        bc_s = self.boundary_params["South"]["type"]
        bc_e = self.boundary_params["East"]["type"]

        self.periodic_we = False
        self.periodic_ns = False

        if bc_n == "Periodic" or bc_s == "Periodic":
            if bc_n == bc_s:
                self.periodic_ns = True
            else:
                raise IOError(
                    'Both "North" and "South" Boundary Conditions need to \
                    be defined as "Periodic" for the periodic treatment to \
                    be applied')

        if bc_w == "Periodic" or bc_e == "Periodic":
            if bc_w == bc_e:
                self.periodic_we = True
            else:
                raise IOError(
                    'Both "West" and "East" Boundary Conditions need to \
                    be defined as "Periodic" for the periodic treatment to \
                    be applied')


def apply_bc(lhs, rhs, metric, boundaries):
    """
    Give the altered version of LHS matrix and RHS vector to apply boundary
    conditions

    Parameters
    ----------
    lhs: left-hand side matrix (A in AX=B)
    rhs: right-hand side matrix (B in AX=B)
    metric: an instance of class Metrics2d containing gradient operators
    boundaries: dictionary with boundary data

    Returns
    -------
    lhs: modified left-hand side matrix
    rhs: modified right-hand side matrix
    """

    for boundary_name, bc_dict in boundaries.items():

        if bc_dict["type"] == 'Robin':
            a_value, b_value, c_value = bc_dict["bc_values"]
            nodes = metric.bnd_nodes[boundary_name]
            grad_n = metric.grad_n_bc
            bc_robin(lhs, rhs, nodes, grad_n, a_value, b_value, c_value)

        elif bc_dict["type"] == 'Neumann':
            # Todo: code Neumann from Robin BC
            pass
        elif bc_dict["type"] == 'Dirichlet':
            dirichlet_value = bc_dict["bc_values"]
            nodes = metric.bnd_nodes[boundary_name]
            grad_n = metric.grad_n_bc
            bc_dirichlet(lhs, rhs, nodes, grad_n, dirichlet_value)
        elif bc_dict["type"] == 'Periodic':
            pass
        else:
            raise NotImplementedError('The only available boundary conditions \
                                      are "Robin", "Neumann", "Dirichlet" and \
                                      "Periodic".')

    return lhs, rhs


# Todo : passer en csr
def bc_robin_csr(matrix_lhs, matrix_rhs, positions, grad_n_bc, a_val,
                 b_val, c_val):
    """
    Function to enforce the Boundary Condition in the Robin formalism:

                    a * (df / dn) + b * f + c = O

    Parameters
    ----------
    matrix_lhs: [CSR] Left Hand Side square matrix
                        (shape (size_i_w * size_j_ns)^2)
    matrix_rhs: [ndarray] Right Hand Side vector
                        (shape (size_i_w * size_j_ns))
    positions: indices of lines corresponding to the patch being processed
    grad_n_bc: [CSR] normal gradient to boundary nodes
    a_val: value of the "a" Robin parameter
    b_val: value of the "b" Robin parameter
    c_val: value of the "a" Robin parameter

    Returns
    -------
    Altered left-hand side matrix and right-hand side vector
    """

    for row in positions:
        # Gradient normal to the boundary if necessary
        if grad_n_bc.count_nonzero():
            new_row = a_val * grad_n_bc.getrow(row).toarray().ravel()
            set_row_csr(matrix_lhs, row, new_row)

        # # Term "b" of robin BC
        diag_bc = np.zeros(matrix_rhs.shape)
        diag_bc[row] = b_val
        mat_bc_b_csr = spsp.csr_matrix(np.diag(diag_bc))
        matrix_lhs = matrix_lhs + mat_bc_b_csr

    # Right hand side for inhomogeneous Dirichlet / Neumann / Robin BCs
    matrix_rhs[positions] = - c_val

    return matrix_lhs, matrix_rhs


def bc_robin(matrix_lhs, matrix_rhs, positions, grad_n_bc, a_val,
             b_val, c_val):
    """
    Function to enforce the Boundary Condition in the Robin formalism:

                    a * (df / dn) + b * f + c = O

    Parameters
    ----------
    matrix_lhs: Left Hand Side square matrix (shape (size_i_w * size_j_ns)^2)
    matrix_rhs: Right Hand Side vector (shape (size_i_w * size_j_ns))
    positions: indices of lines corresponding to the patch being processed
    grad_n_bc: normal gradient to boundary nodes
    a_val: value of the "a" Robin parameter
    b_val: value of the "b" Robin parameter
    c_val: value of the "a" Robin parameter

    Returns
    -------
    Altered left-hand side matrix and right-hand side vector
    """

    for row in positions:
        # Gradient normal to the boundary
        if grad_n_bc.count_nonzero():
            matrix_lhs[row, :] = a_val * grad_n_bc[row, :]

        # # Term "b" of robin BC
        matrix_lhs[row, row] += b_val

    # Right hand side for inhomogeneous Dirichlet / Neumann / Robin BCs
    matrix_rhs[positions] = - c_val

    return matrix_lhs, matrix_rhs


def bc_neumann(matrix_lhs, matrix_rhs, positions, grad_n_bc,
               target_gradient):
    """
    Function to enforce the Neumann Boundary Condition in the Robin formalism:

                    a * (df / dn) + b * f + c = O

    Parameters
    ----------
    grad_n_bc
    target_gradient
    matrix_lhs : Left Hand Side square matrix (shape (size_i_w * size_j_ns)^2)
    matrix_rhs : Right Hand Side vector (shape (size_i_w * size_j_ns))
    positions : indices of lines corresponding to the patch being processed

    Returns
    -------
    Altered left-hand side matrix and right-hand side vector
    """
    lhs_out, rhs_out = bc_robin(matrix_lhs, matrix_rhs, positions,
                                grad_n_bc, 1., 0., -target_gradient)

    return lhs_out, rhs_out


def bc_dirichlet(matrix_lhs, matrix_rhs, positions, grad_n_bc, target_value):
    """
    Function to enforce Dirichlet Boundary Condition in the Robin formalism:

                    a * (df / dn) + b * f + c = O

    Parameters
    ----------
    target_value
    matrix_lhs : Left Hand Side square matrix (shape (size_i_w * size_j_ns)^2)
    matrix_rhs : Right Hand Side vector (shape (size_i_w * size_j_ns))
    positions : indices of lines corresponding to the patch being processed

    Returns
    -------
    Altered left-hand side matrix and right-hand side vector
    """
    print("Dirichlet bc target -->", target_value)

    # lhs_out, rhs_out = bc_robin_csr(matrix_lhs, matrix_rhs, positions,
    #                                 grad_n_bc, 0., 1., -target_value)
    lhs_out, rhs_out = bc_robin(matrix_lhs, matrix_rhs, positions,
                                grad_n_bc, 0., 1., -target_value)

    return lhs_out, rhs_out



"""
This script contains function to properly visualize matrices
"""

import numpy as np


def show_mat(matrix, title, show=False, save=False):
    """
    Show and/or save a matrix visualization.

    Parameters
    ----------
    matrix: 2d matrix
    title: Title of the plot
    show: Boolean to show the plot or not
    save: Boolean to save the plot or nor (automatic name from title)

    Returns
    -------
    None
    """
    import matplotlib.pyplot as plt

    try:
        # If matrix is a csr then take it back to numpy array
        data = matrix.toarray()
    except AttributeError:
        data = np.array(matrix)

    fig = plt.figure()
    plt.title(title)
    axes = fig.add_subplot(111)
    axes.set_xlabel(r"$j_{\rm NS}$")
    axes.set_ylabel(r"$i_{\rm WE}$")
    cax = axes.matshow(data)
    fig.colorbar(cax)
    plt.tight_layout()

    if save:
        name = filter_stupid_characters(title)
        plt.savefig('%s' % name)

    if show:
        plt.show()


def filter_stupid_characters(string):
    """
    Delete and replace stupid characters to save the figure

    Parameters
    ----------
    string: title of the plot to be changed into the filename

    Returns
    -------
    cleaned string
    """
    for char in [",", "$", "\\", " ", "__", "___"]:
        string = string.replace(char, "")

    if string.startswith("_"):
        string = string[1:-1]

    return string



""" module to data array-like dictionnary to files
    for visulaisation or storage pupopses
"""
import imp
import os
import numpy as np
from arnica.utils.nparray2xmf import (create_time_collection_xmf,
                                      NpArray2Xmf)

def dump_dico_2d_nparrays(data_dict, filename,
                          x_coords, y_coords,
                          z_coords, **kw_args):
    """

    Parameters:
    ==========
    data_dict : dictionnary holding the  2d arrays, on the format
                data_dict[key] = array(n1, n2)
                where (n1, n_2) is a subset of (n_x, n_y, n_z)

    filename : the xmf filename

    <x|y|z>_coords : 2d numpy arrays for coordiantes over each axis
                    must be of shape (n_1, n_2)

    keyword args (optional):
    =============
    time  :  physical time corresponding to the array
             used in the xmf file as <Time Value="time"....
    grid_name : the name of the grid to be used in the xmf file
                as <Grid Name="grid_name"....
    domain_name : the name of the domain to be used in the xmf file
                as <Domain Name=domain_name....

    Returns:
    =======
    None
    """
    grid_name = kw_args.get('grid_name')
    time = kw_args.get('time')
    domain_name = kw_args.get('domain_name')

    fields = list(data_dict.keys())


    visu_ = NpArray2Xmf(filename,
                        time=time,
                        mesh_name=grid_name,
                        domain_name=domain_name)
    visu_.create_grid(x_coords, y_coords, z_coords)
    for field in fields:
        visu_.add_field(data_dict[field], field)
    visu_.dump()

def dump_dico_2d_time_nparrays(data_dict, root_path, prefix,
                               x_coords, y_coords, z_coords,
                               **kw_args):
    """ Dumps a dictionnary of time series 2d arrays
       to xmf files

    Parameters:
    ==========
    data_dict : dictionnary holding the times series 2d
                arrays, on the format
                data_dict[key] = array(n_time, n1, n2)
                where:
                    - n_time is the number of time steps
                    - (n1, n_2) is a subset of (n_x, n_y, n_z)

    prefix : the prefix to be used to generate xmf filenames

    <x|y|z>_coords : 2d numpy arrays for coordiantes over each axis
                    must be of shape (n_1, n_2)

    keyword args (optional):
    =============

    steps : a list of integer time series steps
            that will be used to generate xmf files
            on the format : <prefix>_<step>.xmf
            if None steps will be generated as the
            range of time dimension of data arrays

    times : a list of float physical times that will
            be used in xmf files to describe the
            time of each step.
            if None will be generated as the
            range of time dimension of data arrays
    """
    path = os.path.abspath(root_path)
    if not os.path.exists(path):
        os.makedirs(path)

    n_time = data_dict[list(data_dict.keys())[0]].shape[0]

    times = kw_args.get('times')
    steps = kw_args.get('steps')

    if times is None:
        times = np.linspace(0.0, n_time, num=n_time, dtype=float)
    if "steps" is None:
        steps = np.linspace(0, n_time, num=n_time, dtype=int)

    collection = []
    for step in range(n_time):
        step_dict = dict()
        for field in data_dict:
            step_dict[field] = data_dict[field][step, :, :]
        filename = "%s/%s_%08d.xmf" %(path, prefix, steps[step])
        grid_name = "%s_%08d" %(prefix, steps[step])
        dump_dico_2d_nparrays(step_dict, filename, x_coords, y_coords,
                              z_coords, time=times[step], domain_name=prefix,
                              grid_name=grid_name)

        collection.append("%s_%08d.xmf" %(prefix, steps[step]))

    filename = "%s/%s_collection.xmf" %(path, prefix)
    create_time_collection_xmf(collection, filename)


def dump_dico_1d_nparrays(filename, data_dict):
    """Write statistics to file

        Parameters:
        ==========
        filename : the file name to which array dictionnary are dumped
                    possible extensions : - .xlsx (if pandas is found)
                                          - .csv (default format)

        data_dict : a dictionnary holding the data arrays

        Returns:
        =======
        None
    """
    try:
        imp.find_module('pandas')
        import pandas as pd
        pd_found = True
    except ImportError:
        msg = 'dump_dico_1d_nparrays cannot find pandas, the classic csv\
               writing will be used. Pandas is required to dump results to\
               Excel files'
        print(msg)
        pd_found = False

    file_out = os.path.abspath(filename)
    if not os.path.exists(os.path.dirname(file_out)):
        os.makedirs(os.path.dirname(file_out))


    file_out, format_ = os.path.splitext(os.path.abspath(file_out))
    format_ = format_.replace('.', '').lower()

    if format_ not in ["csv", "xlsx"]:
        format_ = 'xlsx'


    if format_ == "xlsx" and pd_found:
        file_out = "%s.%s" %(file_out, format_)
        df_ = pd.DataFrame.from_dict(data_dict)
        df_.to_excel(file_out, index=False)
    else:
        format_ = 'csv'
        file_out = "%s.%s" %(file_out, format_)
        header = " , ".join(list(data_dict.keys()))
        n_columns = len(data_dict.keys())
        n_lines = data_dict[next(iter(data_dict.keys()))].shape[0]
        out_array = np.zeros((n_lines, n_columns))
        for col, key in enumerate(data_dict, 0):
            out_array[:, col] = data_dict[key][:]
        np.savetxt(file_out, out_array, delimiter=' , ', header=header)

def plot_dict_data_as_file(data_dict, filename, x_data, y_data, **kw_args):
    """ Generates and write XY-plot to file

    Parameters:
    ==========
    filename : the file to which the plot is written it contains
                the extension that defines the format e.g: 'plot_toto.png'
                Supported formats/extensions : png, pdf, ps, eps and svg
                If not provided, by default "pdf" extension is used.
    data_dict : a dictionnary holding the data arrays
    x_data : the key to the array holding the abscissa data
    y_data : the key to the array holding the y data

    keyword args (optional):
    =======================
    x_label : label of the x axis, by default x_data is used (supports latex)
    y_label : label of the y axis, by default y_data is used (supports latex)
    """
    import matplotlib.pyplot as plt

    x_label = kw_args.get('x_label')
    if x_label is None:
        x_label = x_data

    y_label = kw_args.get('y_label')
    if y_label is None:
        y_label = y_data

    path = os.path.dirname(os.path.abspath(filename))
    if not os.path.exists(path):
        os.makedirs(path)
    file_out, format_ = os.path.splitext(os.path.abspath(filename))
    format_ = format_.replace('.', '').lower()

    if format_ not in ["png", "pdf", "ps", "eps", "svg"]:
        format_ = 'pdf'


    plt.rc('text', usetex=True)
    plt.rc('font', family='serif')
    fig = plt.figure(figsize=(5.5, 5))
    fig.subplots_adjust(left=0.15, right=0.97, bottom=0.15,
                        top=0.9, wspace=0.27)

    axe = fig.add_subplot(111)
    axe.plot(data_dict[x_data], data_dict[y_data], c='k', lw=2, marker='s')
    axe.set_ylabel(y_label, fontsize=20, fontweight='bold')
    axe.set_xlabel(x_label, fontsize=20, fontweight='bold')
    axe.grid(which='major', color='gray', linestyle='--',
             dashes=(8, 12), linewidth=0.5)

    axe.minorticks_on()

    axe.ticklabel_format(axis='both', style='sci', scilimits=(0, 0))

    for axis in [axe.xaxis, axe.yaxis]:
        for tick in axis.get_major_ticks():
            tick.label.set_fontsize(16)
        tick = axis.get_offset_text()
        tick.set_size(16)

    file_out = "%s.%s" %(file_out, format_)
    plt.savefig(file_out)



"""
This module is contains the core of Finite Differences 2D solvers.

The class metrics:
    1) Builds the jacobian to handle curvilinear meshes.
    2) Computes the gradients operators
    3) Computes the boundary-normal-gradient operator
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import scipy.sparse as scp

from arnica.utils.nparray2xmf import NpArray2Xmf

WRITE_BNDY = False


# Todo : think about axi-periodic treatments
class Metrics2d:
    """
    Compute operators onto a curvilinear 2D mesh
    """

    def __init__(self, x_coor, y_coor, periodic_ns=False, periodic_we=False):
        """
        Metric2d constructor

        Parameters
        ----------
        x_coor: x_coordinates as a numpy meshgrid
        y_coor: y_coordinates as a numpy meshgrid
        periodic_ns: Boolean, True if periodic North-South BC
        periodic_we: Boolean, True if periodic West-East BC
        """
        self.x_coor = x_coor
        self.y_coor = y_coor

        self.size_we, self.size_ns = x_coor.shape
        self.size = self.size_we * self.size_ns

        self.shp2d = x_coor.shape
        self.shp1d = x_coor.ravel().shape

        self.periodic_ns = periodic_ns
        self.periodic_we = periodic_we

        print('\n\t\t Metric:\t _compute_metric()')
        self._compute_metric()

        print("\n\t\t Metric:\t  _compute_matrices_csr()")
        self._compute_matrices_csr()

        print("\n\t\t Metric:\t  _get_bc_iterators()")
        self._get_bc_iterators()

        print("\n\t\t Metric:\t  _compute_normals()")
        self._compute_normals()

    def _compute_metric(self):
        """
        Compute the metric elements
        """
        diff_x_ns = self.diff_ns_2(self.x_coor)
        diff_x_we = self.diff_we_2(self.x_coor)
        diff_y_ns = self.diff_ns_2(self.y_coor)
        diff_y_we = self.diff_we_2(self.y_coor)
        inv_det_j = np.reciprocal(
            diff_x_ns * diff_y_we - diff_x_we * diff_y_ns)

        self.jacob_x_ns = inv_det_j * diff_x_ns
        self.jacob_x_we = inv_det_j * diff_x_we
        self.jacob_y_ns = inv_det_j * diff_y_ns
        self.jacob_y_we = inv_det_j * diff_y_we

    def check_compute_matrices(self):
        """
        compute the matrix form of operators
        DO NOT REMOVE! --> USED FOR TESTING CSR GRADIENTS
        """

        def ijc(i_idx, j_idx):
            """ Give indexes in 1D version """
            return j_idx + self.size_ns * i_idx

        def iters(k, size_k, perio):
            """Provide iterators, taking boundaries into account """

            cfk = 0.5
            km1 = k - 1
            kp1 = k + 1
            if k == 0:
                if perio:
                    km1 = size_k - 1
                else:
                    cfk = 1.
                    km1 = 0

            if k == size_k - 1:
                if perio:
                    kp1 = 0
                else:
                    cfk = 1.
                    kp1 = size_k - 1

            return km1, kp1, cfk

        mat_gx = np.zeros((self.size, self.size))
        mat_gy = np.zeros((self.size, self.size))

        for i_we in range(self.size_we):
            im1, ip1, cfi = iters(i_we, self.size_we, self.periodic_ns)
            for j_ns in range(self.size_ns):
                jm1, jp1, cfj = iters(j_ns, self.size_ns, self.periodic_we)

                jcb_y_we = self.jacob_y_we[i_we, j_ns]
                jcb_y_ns = self.jacob_y_ns[i_we, j_ns]
                jcb_x_we = self.jacob_x_we[i_we, j_ns]
                jcb_x_ns = self.jacob_x_ns[i_we, j_ns]

                mat_gx[ijc(i_we, j_ns), ijc(i_we, jp1)] += cfj * jcb_y_we
                mat_gx[ijc(i_we, j_ns), ijc(i_we, jm1)] -= cfj * jcb_y_we
                mat_gx[ijc(i_we, j_ns), ijc(ip1, j_ns)] -= cfi * jcb_y_ns
                mat_gx[ijc(i_we, j_ns), ijc(im1, j_ns)] += cfi * jcb_y_ns

                mat_gy[ijc(i_we, j_ns), ijc(i_we, jp1)] -= cfj * jcb_x_we
                mat_gy[ijc(i_we, j_ns), ijc(i_we, jm1)] += cfj * jcb_x_we
                mat_gy[ijc(i_we, j_ns), ijc(ip1, j_ns)] += cfi * jcb_x_ns
                mat_gy[ijc(i_we, j_ns), ijc(im1, j_ns)] -= cfi * jcb_x_ns

        self.grad_x_slow = scp.csr_matrix(mat_gx)
        self.grad_y_slow = scp.csr_matrix(mat_gy)

    def _compute_matrices_csr(self):
        """ compute the matrix form of operators """
        size = self.size_we * self.size_ns
        self.mat_gx = np.zeros((size, size))
        self.mat_gy = np.zeros((size, size))

        # Option 1
        ijc_array = np.arange(self.size_we * self.size_ns)
        iters_i = np.repeat(np.arange(self.size_we), self.size_ns)
        iters_j = np.tile(np.arange(self.size_ns), self.size_we)
        ip1 = np.arange(1, self.size_we + 1)
        jp1 = np.arange(1, self.size_ns + 1)
        im1 = np.arange(-1, self.size_we - 1)
        jm1 = np.arange(-1, self.size_ns - 1)

        cfi_sb = np.ones(self.size_we) * 0.5
        cfj_sb = np.ones(self.size_ns) * 0.5
        if self.periodic_ns:
            im1[0], ip1[-1] = self.size_we - 1, 0
        else:
            im1[0], ip1[-1] = 0, self.size_we - 1
            cfi_sb[0], cfi_sb[-1] = 1., 1.

        if self.periodic_we:
            jm1[0], jp1[-1] = self.size_ns - 1, 0
        else:
            jm1[0], jp1[-1] = 0, self.size_ns - 1
            cfj_sb[0], cfj_sb[-1] = 1., 1.

        cfi = np.repeat(cfi_sb, self.size_ns)
        cfj = np.tile(cfj_sb, self.size_we)

        iters_ip1 = np.repeat(ip1, self.size_ns)
        iters_im1 = np.repeat(im1, self.size_ns)
        iters_jp1 = np.tile(jp1, self.size_we)
        iters_jm1 = np.tile(jm1, self.size_we)
        point_jp1 = ijc_array[iters_i * self.size_ns + iters_jp1]
        point_jm1 = ijc_array[iters_i * self.size_ns + iters_jm1]
        point_ip1 = ijc_array[iters_j + iters_ip1 * self.size_ns]
        point_im1 = ijc_array[iters_j + iters_im1 * self.size_ns]

        column_x = np.concatenate(
            (
                point_jp1[:, None],
                point_jm1[:, None],
                point_ip1[:, None],
                point_im1[:, None],
            ),
            axis=1,
        ).ravel()
        column_y = np.concatenate(
            (
                point_ip1[:, None],
                point_im1[:, None],
                point_jp1[:, None],
                point_jm1[:, None],
            ),
            axis=1,
        ).ravel()
        array_x = np.concatenate(
            (
                (cfj * self.jacob_y_we.ravel())[:, None],
                (-cfj * self.jacob_y_we.ravel())[:, None],
                (-cfi * self.jacob_y_ns.ravel())[:, None],
                (cfi * self.jacob_y_ns.ravel())[:, None],
            ),
            axis=1,
        ).ravel()
        array_y = np.concatenate(
            (
                (cfi * self.jacob_x_ns.ravel())[:, None],
                (-cfi * self.jacob_x_ns.ravel())[:, None],
                (-cfj * self.jacob_x_we.ravel())[:, None],
                (cfj * self.jacob_x_we.ravel())[:, None],
            ),
            axis=1,
        ).ravel()

        grad_x = scp.csr_matrix(
            (array_x, (np.repeat(ijc_array, 4), column_x)),
            shape=(size, size)
        )
        grad_y = scp.csr_matrix(
            (array_y, (np.repeat(ijc_array, 4), column_y)),
            shape=(size, size)
        )
        self.lapl = (grad_x.dot(grad_x) + grad_y.dot(grad_y))  # .tolil()
        self.grad_x_csr = grad_x
        self.grad_y_csr = grad_y

    def _compute_normals(self):
        """
        Compute unit normal vector over the boundaries
        Parameters
        ----------
        Arrays of unit normal vectors : nx and ny.
        """

        x_1d = self.x_coor.ravel()
        y_1d = self.y_coor.ravel()

        normal = np.zeros((self.size_we * self.size_ns, 2))

        # Todo: south and South corners overwritten by east and west
        for bnd_name in ["North", "South", "West", "East"]:
            idx = self.bnd_nodes[bnd_name]

            # Compute dx and dy
            dx_bnd = np.diff(x_1d[idx])
            dy_bnd = np.diff(y_1d[idx])

            # Todo: moyenne artihmetique des normales aux faces adjacentes afin d'avoir les normales aux noeuds + trick perio (moyenne arithmetique avec avec noeud periodique)
            # Compute side effect of diff function for dx and dy
            dx_bnd_end = np.diff(np.roll(x_1d[idx], 1))[-1]
            dy_bnd_end = np.diff(np.roll(y_1d[idx], 1))[-1]

            # Fill normal matrix
            normal[idx[0:-1], 0] = dy_bnd
            normal[idx[-1], 0] = dy_bnd_end
            normal[idx[0:-1], 1] = -dx_bnd
            normal[idx[-1], 1] = -dx_bnd_end

            # Normalize each normal vector
            norm_axis1 = np.linalg.norm(normal[idx, :], axis=1)
            normal[idx] /= norm_axis1.repeat(2).reshape(normal[idx, :].shape)

        # Reshape unit normal vector for output
        print("\n\t\t Metric:\t _compute_normals()\treshape")
        normal_2d_x = normal[:, 0].reshape((self.size_we, self.size_ns))
        normal_2d_y = normal[:, 1].reshape((self.size_we, self.size_ns))

        size_2d = self.size_we, self.size_ns
        if WRITE_BNDY:
            print('\n\t--> Writing "Boundaries.h5"')
            out_h5 = NpArray2Xmf("Boundaries.h5")
            out_h5.create_grid(self.x_coor, self.y_coor, np.zeros(size_2d))

            for bnd_name in ["North", "South", "West", "East"]:
                idx = self.bnd_nodes[bnd_name]
                bnd = np.zeros_like(self.x_coor).ravel()
                bnd[idx] = 1
                out_h5.add_field(bnd.reshape(size_2d), "BC_%s" % bnd_name)

            out_h5.add_field(normal_2d_x, "nx")
            out_h5.add_field(normal_2d_y, "ny")
            out_h5.dump()

        print("\n\t\t Metric:\t _compute_normals()\tstore normals")
        self.n_x = scp.csr_matrix(np.diag(normal_2d_x.ravel()))
        self.n_y = scp.csr_matrix(np.diag(normal_2d_y.ravel()))

        # Gradient normal to the boundary

        # nx_grad_x = np.dot(np.diag(normal_2d_x.ravel()), self.grad_x)
        # ny_grad_y = np.dot(np.diag(normal_2d_y.ravel()), self.grad_y)
        print("\n\t\t Metric:\t _compute_normals()\tcompute nx grad x")
        nx_grad_x = self.n_x.dot(self.grad_x_csr)
        ny_grad_y = self.n_y.dot(self.grad_y_csr)

        print("\n\t\t Metric:\t _compute_normals()\tcompute n grad n")
        self.grad_n_bc = nx_grad_x + ny_grad_y

    def _get_bc_iterators(self):
        """
        Gives node number of boundary patches
        """

        # Flip of numpy array necessary to compute the normals
        i_we = self.size_we - 1
        north = np.flip(
            np.array([(j_idx + self.size_ns * i_we) for j_idx in
                      range(self.size_ns)]))

        # Flip of numpy array necessary to compute the normals
        j_ns = 0
        west = np.flip(
            np.array([(j_ns + self.size_ns * i_idx) for i_idx in
                      range(self.size_we)]))

        i_we = 0
        south = np.array(
            [(j_idx + self.size_ns * i_we) for j_idx in range(self.size_ns)])

        j_ns = self.size_ns - 1
        east = np.array(
            [(j_ns + self.size_ns * i_idx) for i_idx in range(self.size_we)])

        self.bnd_nodes = {}
        self.bnd_nodes["North"] = north
        self.bnd_nodes["West"] = west
        self.bnd_nodes["South"] = south
        self.bnd_nodes["East"] = east

    def diff_ns_2(self, array_):
        """ compute differentiation in direction west (j-direction)"""
        out = np.zeros_like(array_)
        size_j = out.shape[1]

        out[:, range(1, size_j - 1)] = 0.5 * (array_[:, range(2, size_j)]
                                              - array_[:,
                                                range(0, size_j - 2)])

        if self.periodic_we:
            out[:, 0] = 0.5 * (array_[:, 1] - array_[:, size_j - 1])

            out[:, size_j - 1] = 0.5 * (array_[:, 0] - array_[:, size_j - 2])
        else:
            out[:, 0] = array_[:, 1] - array_[:, 0]
            out[:, size_j - 1] = array_[:, size_j - 1] - array_[:, size_j - 2]

        return out

    def diff_we_2(self, array_):
        """ compute differentiation in direction eta (i-direction)"""
        out = np.zeros_like(array_)
        size_i = out.shape[0]

        out[range(1, size_i - 1), :] = 0.5 * (array_[range(2, size_i), :]
                                              - array_[range(0, size_i - 2),
                                                :])

        if self.periodic_ns:
            out[0, :] = 0.5 * (array_[1, :] - array_[size_i - 1, :])

            out[size_i - 1, :] = 0.5 * (array_[0, :] - array_[size_i - 2, :])
        else:
            out[0, :] = array_[1, :] - array_[0, :]
            out[size_i - 1, :] = array_[size_i - 1, :] - array_[size_i - 2, :]

        return out


def set_row_csr(csr, row_idx, new_row):
    """
    Replace a row in a CSR sparse matrix A.

    Parameters
    ----------
    csr: csr_matrix
        Matrix to change
    row_idx: int
        index of the row to be changed
    new_row: np.array
        list of new values for the row of A

    Returns
    -------
    None (the matrix A is changed in place)

    Prerequisites
    -------------
    The row index shall be smaller than the number of rows in A
    The number of elements in new row must be equal to the number of columns in
    matrix A
    """
    assert scp.isspmatrix_csr(csr), 'A shall be a csr_matrix'
    assert row_idx < csr.shape[0], \
        'The row index ({0}) shall be smaller than the number of rows in A ({1})' \
            .format(row_idx, csr.shape[0])
    try:
        n_elements_new_row = len(new_row)
    except TypeError:
        msg = 'Argument new_row shall be a list or numpy array, is now a {0}' \
            .format(type(new_row))
        raise AssertionError(msg)
    n_cols = csr.shape[1]
    assert n_cols == n_elements_new_row, \
        'The number of elements in new row ({0}) must be equal to ' \
        'the number of columns in matrix A ({1})' \
            .format(n_elements_new_row, n_cols)

    idx_start = csr.indptr[row_idx]
    idx_end = csr.indptr[row_idx + 1]
    additional_nnz = n_cols - (idx_end - idx_start)

    # Substitute dense data
    csr.data = np.r_[csr.data[:idx_start], new_row, csr.data[idx_end:]]

    # Correct indices
    csr.indices = np.r_[csr.indices[:idx_start],
                        np.arange(n_cols),
                        csr.indices[idx_end:]]

    # Correct indptr
    csr.indptr = np.r_[csr.indptr[:row_idx + 1],
                       csr.indptr[(row_idx + 1):]
                       + additional_nnz]



""" Module to test the heat transfer
resolution on grids curvilinear 2D grids """

from __future__ import absolute_import, division
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as sciopt
import scipy.sparse as scp
from arnica.utils.nparray2xmf import NpArray2Xmf


class Metrics(object):
    """ metrics class """

    def __init__(self, x_coor, y_coor):
        """ startup class
        x_coor : numpy array (n,m) , x_coordinates
        y_coor : numpy array (n,m) , y_coordinates
        """
        self.x_coor = x_coor
        self.y_coor = y_coor

        self.size_i, self.size_j = x_coor.shape

        self.shp2d = x_coor.shape
        self.shp1d = x_coor.ravel().shape

        self.order = 2
        self.periodic_xi = False
        self.periodic_eta = False
        self._compute_metric()
        # self._compute_matrixes()
        self._compute_matrixes_csr()

    def _compute_metric(self):
        """ compute the metric elements """
        diff_x_xi = self.diff_xi2(self.x_coor)
        diff_x_eta = self.diff_eta2(self.x_coor)
        diff_y_xi = self.diff_xi2(self.y_coor)
        diff_y_eta = self.diff_eta2(self.y_coor)
        inv_det_j = np.reciprocal(diff_x_xi*diff_y_eta
                                  - diff_x_eta*diff_y_xi)

        self.jacob_x_xi = inv_det_j*diff_x_xi
        self.jacob_x_eta = inv_det_j*diff_x_eta
        self.jacob_y_xi = inv_det_j*diff_y_xi
        self.jacob_y_eta = inv_det_j*diff_y_eta

    # def _compute_matrixes(self):
    #     """ compute the matrix form of operatoirs """
    #     size = self.size_i * self.size_j
    #     self.mat_gx = np.zeros((size, size))
    #     self.mat_gy = np.zeros((size, size))

    #     def ijc(i, j):
    #         """ Give indexes in 1D version """
    #         return j + self.size_j*i

    #     def iters(k, size_k, perio, order=1):
    #         """Provide terators , taking boundaries into account """
    #         if self.order == 1:
    #             cfk = 1.
    #             km1 = int(k)
    #             kp1 = k+1
    #             if k == size_k-1:
    #                 if perio:
    #                     kp1 = 0
    #                 else:
    #                     km1 = size_k-2
    #                     kp1 = size_k-1

    #         if self.order == 2:
    #             cfk = 0.5
    #             km1 = k-1
    #             kp1 = k+1
    #             if k == 0:
    #                 if perio:
    #                     km1 = size_k-1
    #                 else:
    #                     cfk = 1.
    #                     km1 = 0

    #             if k == size_k-1:
    #                 if perio:
    #                     kp1 = 0
    #                 else:
    #                     cfk = 1.
    #                     kp1 = size_k-1

    #         return km1, kp1, cfk

    #     for i in range(self.size_i):
    #         im1, ip1, cfi = iters(i, self.size_i, self.periodic_eta)
    #         for j in range(self.size_j):
    #             jm1, jp1, cfj = iters(j, self.size_j, self.periodic_xi)
    #             self.mat_gx[ijc(i, j), ijc(i, jp1)] += cfj*self.jacob_y_eta[i, j]
    #             self.mat_gx[ijc(i, j), ijc(i, jm1)] += -cfj*self.jacob_y_eta[i, j]
    #             self.mat_gx[ijc(i, j), ijc(ip1, j)] += -cfi*self.jacob_y_xi[i, j]
    #             self.mat_gx[ijc(i, j), ijc(im1, j)] += +cfi*self.jacob_y_xi[i, j]

    #             self.mat_gy[ijc(i, j), ijc(ip1, j)] += cfi*self.jacob_x_xi[i, j]
    #             self.mat_gy[ijc(i, j), ijc(im1, j)] += -cfi*self.jacob_x_xi[i, j]
    #             self.mat_gy[ijc(i, j), ijc(i, jp1)] += -cfj*self.jacob_x_eta[i, j]
    #             self.mat_gy[ijc(i, j), ijc(i, jm1)] += +cfj*self.jacob_x_eta[i, j]

    #     self.matlap = np.dot(self.mat_gx, self.mat_gx) + np.dot(self.mat_gy, self.mat_gy)

    def _compute_matrixes_csr(self):
        """ compute the matrix form of operators """
        size = self.size_i * self.size_j
        self.mat_gx = np.zeros((size, size))
        self.mat_gy = np.zeros((size, size))

        # Option 1
        ijc_array = np.arange(self.size_i * self.size_j)
        iters_i = np.repeat(np.arange(self.size_i), self.size_j)
        iters_j = np.tile(np.arange(self.size_j), self.size_i)
        ip1 = np.arange(1, self.size_i + 1)
        jp1 = np.arange(1, self.size_j + 1)
        im1 = np.arange(-1, self.size_i - 1)
        jm1 = np.arange(-1, self.size_j - 1)

        cfi_sb = np.ones(self.size_i) * 0.5
        cfj_sb = np.ones(self.size_j) * 0.5
        if self.periodic_xi:
            im1[0], ip1[-1] = self.size_i - 1, 0
        else:
            im1[0], ip1[-1] = 0, self.size_i - 1
            cfi_sb[0], cfi_sb[-1] = 1., 1.

        if self.periodic_eta:
            jm1[0], jp1[-1] = self.size_j - 1, 0
        else:
            jm1[0], jp1[-1] = 0, self.size_j - 1
            cfj_sb[0], cfj_sb[-1] = 1., 1.

        cfi = np.repeat(cfi_sb, self.size_j)
        cfj = np.tile(cfj_sb, self.size_i)

        iters_ip1 = np.repeat(ip1, self.size_j)
        iters_im1 = np.repeat(im1, self.size_j)
        iters_jp1 = np.tile(jp1, self.size_i)
        iters_jm1 = np.tile(jm1, self.size_i)
        point_jp1 = ijc_array[iters_i * self.size_i + iters_jp1]
        point_jm1 = ijc_array[iters_i * self.size_i + iters_jm1]
        point_ip1 = ijc_array[iters_j + iters_ip1 * self.size_i]
        point_im1 = ijc_array[iters_j + iters_im1 * self.size_i]

        column_x = np.concatenate(
            (
                point_jp1[:, None],
                point_jm1[:, None],
                point_ip1[:, None],
                point_im1[:, None],
            ),
            axis=1,
        ).ravel()
        column_y = np.concatenate(
            (
                point_ip1[:, None],
                point_im1[:, None],
                point_jp1[:, None],
                point_jm1[:, None],
            ),
            axis=1,
        ).ravel()
        array_x = np.concatenate(
            (
                (cfj * self.jacob_y_eta.ravel())[:, None],
                (-cfj * self.jacob_y_eta.ravel())[:, None],
                (-cfi * self.jacob_y_xi.ravel())[:, None],
                (cfi * self.jacob_y_xi.ravel())[:, None],
            ),
            axis=1,
        ).ravel()
        array_y = np.concatenate(
            (
                (cfi * self.jacob_x_xi.ravel())[:, None],
                (-cfi * self.jacob_x_xi.ravel())[:, None],
                (-cfj * self.jacob_x_eta.ravel())[:, None],
                (cfj * self.jacob_x_eta.ravel())[:, None],
            ),
            axis=1,
        ).ravel()

        grad_x = scp.csr_matrix(
            (array_x, (np.repeat(ijc_array, 4), column_x)),
            shape=(size, size)
        )
        grad_y = scp.csr_matrix(
            (array_y, (np.repeat(ijc_array, 4), column_y)),
            shape=(size, size)
        )
        self.lapl = (grad_x.dot(grad_x) + grad_y.dot(grad_y)) # .tolil()

    def diff_xi2(self, array_):
        """ compute differenciation in direction xi (j-direction)"""
        out = np.zeros_like(array_)
        size_j = out.shape[1]

        if self.order == 1:
            out[:, range(0, size_j - 1)] = (
                array_[:, range(1, size_j)] - array_[:, range(0, size_j - 1)])

            if self.periodic_xi:
                out[:, size_j-1] = (
                    array_[:, 0] - array_[:, size_j - 1])
            else:
                out[:, size_j-1] = array_[:, size_j-1] - array_[:, size_j-2]

        if self.order == 2:
            out[:, range(1, size_j - 1)] = 0.5 * (
                array_[:, range(2, size_j)] - array_[:, range(0, size_j - 2)])

            if self.periodic_xi:
                out[:, 0] = 0.5 * (
                    array_[:, 1] - array_[:, size_j-1])

                out[:, size_j-1] = 0.5 * (
                    array_[:, 0] - array_[:, size_j-2])
            else:
                out[:, 0] = array_[:, 1] - array_[:, 0]
                out[:, size_j-1] = array_[:, size_j-1] - array_[:, size_j-2]

        return out

    def diff_eta2(self, array_):
        """ compute differenciation in direction eta (i-direction)"""
        out = np.zeros_like(array_)
        size_i = out.shape[0]

        if self.order == 1:
            out[range(0, size_i-1), :] = (
                array_[range(1, size_i), :] - array_[range(0, size_i-1), :])

            if self.periodic_eta:
                out[size_i-1, :] = (
                    array_[0, :] - array_[size_i-1, :])
            else:
                out[size_i-1, :] = array_[size_i-1, :] - array_[size_i-2, :]

        if self.order == 2:
            out[range(1, size_i-1), :] = 0.5 * (
                array_[range(2, size_i), :] - array_[range(0, size_i-2), :])

            if self.periodic_eta:
                out[0, :] = 0.5 * (
                    array_[1, :] - array_[size_i-1, :])

                out[size_i-1, :] = 0.5 * (
                    array_[0, :] - array_[size_i-2, :])
            else:
                out[0, :] = array_[1, :] - array_[0, :]
                out[size_i-1, :] = array_[size_i-1, :] - array_[size_i-2, :]
        return out

    def compute_gradient(self, field):
        """ Compute the gradient of a field

        Parameters :
        -----------
        field : numpy array (n,m) , field

        Returns :
        ---------
        (gradient_x  gradient_y),: set of gradients in each direction

        """
        diff_f_xi = self.diff_xi2(field)
        diff_f_eta = self.diff_eta2(field)
        grad_x = diff_f_xi*self.jacob_y_eta - diff_f_eta*self.jacob_y_xi
        grad_y = diff_f_eta*self.jacob_x_xi - diff_f_xi*self.jacob_x_eta
        return grad_x, grad_y

    def compute_laplacian(self, field):
        """ Compute the diagonal elements of the hessian of a field

        Parameters :
        -----------
        field : numpy array (n,m) , field

        Returns :
        ---------
        (hess_xx  hess_yy),:  hessiang diagonal for each direction

        """
        #grad_x, grad_y = self.compute_gradient(field)
        laplacian = ((self.diff_xi2(self.diff_xi2(field)
                                    * self.jacob_y_eta
                                    - self.diff_eta2(field)
                                    * self.jacob_y_xi) * self.jacob_y_eta
                      - self.diff_eta2(self.diff_xi2(field)
                                       * self.jacob_y_eta
                                       - self.diff_eta2(field)
                                       * self.jacob_y_xi) * self.jacob_y_xi)
                     + (self.diff_eta2(self.diff_eta2(field)
                                       * self.jacob_x_xi
                                       - self.diff_xi2(field)
                                       * self.jacob_x_eta)*self.jacob_x_xi
                        - self.diff_xi2(self.diff_eta2(field)
                                        * self.jacob_x_xi
                                        - self.diff_xi2(field)
                                        * self.jacob_x_eta)*self.jacob_x_eta))
        return laplacian


def gen_cart_grid(gridrange,
                  gridpoints):
    """ Generate cartesian grid.

    Parameters :
    ------------
    gridrange : tuple of foats, dimensions of the grid
    gridpoints : tuple of ints (n,m), sampling on the grid

    Returns :
    ---------

    x_coor, y_coor : numpy arrays (n,m) with coordiantes
    """
    x_vec = np.linspace(0, gridrange[0], gridpoints[0], endpoint=True)
    y_vec = np.linspace(0, gridrange[1], gridpoints[1], endpoint=True)
    x_coor, y_coor = np.meshgrid(x_vec, y_vec)
    return x_coor, y_coor


def dilate_center(x_coor, y_coor, perturbation=1.0):
    """ pertubate cartesian mesh dilatation in the center

    Parameters :
    -----------
    x_coor : numpy array (n,m) , x_coordinates
    y_coor : numpy array (n,m) , y_coordinates
    perturbation : float, amplitude of the perturbation perturbation
                   with respect to the grid size

    Returns :
    ----------
    x_coor : numpy array (n,m) , x_coordinates shifted
    y_coor : numpy array (n,m) , y_coordinates shifted

     """
    x_width = x_coor.max() - x_coor.min()
    y_width = y_coor.max() - y_coor.min()
    x_center = x_coor.min() + 0.5*x_width
    y_center = y_coor.min() + 0.5*y_width

    adim_rad = np.sqrt(((x_coor-x_center)/(0.5*x_width))**2
                       + ((y_coor-y_center)/(0.5*y_width))**2)

    rad_dir_x = (x_coor-x_center) / np.maximum(adim_rad, 1.e-16)
    rad_dir_y = (y_coor-y_center) / np.maximum(adim_rad, 1.e-16)
    dilate = np.maximum((1.-adim_rad), 0)
    dilate = 3*dilate*dilate*(1-dilate)

    shift_x = perturbation*x_width*dilate*rad_dir_x
    shift_y = perturbation*y_width*dilate*rad_dir_y

    return x_coor+shift_x, y_coor+shift_y, dilate


def donut_function(x_coor, y_coor, center, radius):
    """
    create a donut test function, with analytically known derivatives

    Parameters :
    -----------
    x_coor : numpy array (n,m) , x_coordinates
    y_coor : numpy array (n,m) , y_coordinates
    center : tuple of floats, center
    radius ; float, radius of donut

    Returns :
    ---------
    donut : numpy array (n,m) , donut test function

    """
    r_coor = np.hypot(x_coor-center[0], y_coor-center[1])
    r_calib = (np.clip(r_coor/radius, 0.5, 1.0)*2-1)
    donut = np.cos(r_calib*np.pi)

    mslope = np.pi/(0.5*radius)
    diff_donut = np.sin(r_calib*np.pi)*mslope
    diff_donut = np.where(r_calib > 0, diff_donut, 0)
    diff_donut = np.where(r_calib < 1, diff_donut, 0)

    diff2_donut = -np.cos(r_calib*np.pi)*mslope*mslope
    diff2_donut = np.where(r_calib > 0, diff2_donut, 0)
    diff2_donut = np.where(r_calib < 1, diff2_donut, 0)

    print('Max grad, grad2 ', mslope, mslope**2)

    return donut, diff_donut, diff2_donut


def double_sine_function(x_coor, y_coor):
    """
    create a  double_sine function, with analytically known derivatives

    Parameters :
    -----------
    x_coor : numpy array (n,m), x_coordinates
    y_coor : numpy array (n,m), y_coordinates

    Returns :
    ---------
    slop : numpy array (n,m) , donut test function

    """
    return np.sin(x_coor)*np.sin(y_coor)


def slope_function(coor, min_val, max_val):
    """
    create a slope x test function, with analytically known derivatives

    Parameters :
    -----------
    coor : numpy array (n,m) , coordinates
    min_val : minimum value
    max_val ; maximum value

    Returns :
    ---------
    slop : numpy array (n,m) , slope test function

    """

    min_c = coor.min()
    max_c = coor.max()

    out = np.ones_like(coor)*min_val
    out += (coor-min_c)/(max_c-min_c) * (max_val-min_val)

    return out


def compute_constants(params):
    """ compute_constants parameters

    Parameters :
    ------------
    params : dict of setup parameters


    Returns :
    ------------
    const : dict of constant parameters
    """

    const = {}
    const["h_ratio"] = params["h_cold"]/params["h_hot"]
    const["biot"] = params["h_hot"] * params["width"]/params["lambda"]
    const["t_ref"] = (params["t_cold"]
                      + ((params["t_hot"] - params["t_cold"])
                         * (1. / (1. + const["h_ratio"]))))
    const["dx"] = params["typ_size"] / params["res"]

    const["lam_ov_rhocp"] = params["lambda"]/(params["rho"]*params["cp"])

    const["dt"] = params["fourier"] * const["dx"]**2 / const["lam_ov_rhocp"]
    const["total_time"] = params["typ_size"]**2 / const["lam_ov_rhocp"]
    const["exp_iterations"] = const["total_time"] / const["dt"]

    for key in const:
        print(key
              + ":"
              + str(const[key]))

    return const


def explicit_solve(temp0, metric, params, const, t_ref):
    """ Explicit time marching solver """

    def residual_max(res):
        """ Compute the non dimensional residual maxx """
        return (np.abs(res).max()
                / (params["t_hot"]-params["t_cold"])
                * const["total_time"]/const["dt"])

    def gap(temp):
        """ return the maximum absolute error """
        return np.abs(t_ref-temp).max()

    loop = 0
    residual = temp0
    temp = temp0
    counter = 0
    while gap(temp) > params["tol"] and loop < params["max_loop"]:
        loop += 1
        counter += 1
        temp[0, :] = params["t_cold"]
        temp[-1, :] = params["t_hot"]
        residual = (const["dt"]
                    * const["lam_ov_rhocp"]
                    * metric.compute_laplacian(temp))
        temp += residual

        if counter > 1000:
            counter = 0
            print("iterations :" + str(loop))
            print("relative residual : " + str(residual_max(residual)))

    print("final time : " + str(loop*const["dt"]))

    return temp


def optimize_solve(temp0, metric, params, const):
    """ give solution by root finding """
    shape2d = (params["res"], params["res"])

    def res_func(temp):
        """ test function for fsolve"""
        rhs2d = (const["dt"]
                 * const["lam_ov_rhocp"]
                 * metric.compute_laplacian(temp.reshape(shape2d)))

        rhs2d[:, 0] = params["t_cold"] - temp.reshape(shape2d)[:, 0]
        rhs2d[:, -1] = params["t_hot"] - temp.reshape(shape2d)[:, -1]

        return rhs2d.ravel()
    # root
    #opt = sciopt.root(res_func, temp0.ravel())
    # print opt.success
    #out = opt.x.reshape(shape2d)

    out = sciopt.fmin_bfgs(res_func, temp0.ravel())
    return out


def showmat(matrx):
    """ show matrix """
    fig = plt.figure()
    axes = fig.add_subplot(111)
    cax = axes.matshow(matrx)
    fig.colorbar(cax)
    plt.show()

def apply_bc_csr(
        matrix_csr_operator,
        size_i,
        size_j,
        min_val,
        max_val):
    """ give the altered version of matrix and RHS to apply CLS """
    i = 0
    bnd_south = np.array([(j + size_j*i) for j in range(size_j)])
    i = size_i-1
    bnd_north = np.array([(j + size_j*i) for j in range(size_j)])
    # j = 0
    # bnd_west = [(j + size_j*i) for i in range(size_i)]
    # j = size_j-1
    # bnd_east = [(j + size_j*i) for i in range(size_i)]

    matrix_lhs = scp.csr_matrix.copy(matrix_csr_operator)
    rhs = np.zeros(size_i * size_j)
    def enforce_bc_dirichlet(positions, value):
        """ single function to enforce the bc effect """
        for row in positions:
            mat_bc.data[mat_bc.indptr[row]:mat_bc.indptr[row+1]] = 0.
        mat_bc[positions, positions] = 1.
        # for pos in positions:
        #    mat_bc[pos, :] = 0.
        #    mat_bc[pos, pos] = 1.
        # initial version for non-csr matrix:
        #mat_bc[positions, :] = 0.
        #mat_bc[positions, pos] = 1.
        rhs[positions] = value

    enforce_bc_dirichlet(bnd_north, max_val)
    #enforce_bc(bnd_west, max_val)
    enforce_bc_dirichlet(bnd_south, min_val)
    #enforce_bc(bnd_east, min_val)
    return mat_bc, rhs


def main_2dheattreansf(params):
    """ Set up main 2D theat transfert resolution

    Parameters :
    ------------
    params : dict of setup parameters

    """
    const = compute_constants(params)
    x_coor, y_coor = gen_cart_grid(gridrange=(params["typ_size"],
                                              params["typ_size"]),
                                   gridpoints=(params["res"],
                                               params["res"]))
    if params["dilate_grid"]:
        x_coor, y_coor, _ = dilate_center(x_coor, y_coor)

    metric = Metrics(x_coor, y_coor)
    #temp0 = np.ones_like(x_coor)

    temp_final = slope_function(y_coor, params["t_cold"], params["t_hot"])
    #temp0 = double_sine_function(x_coor, y_coor)
    #lap2 = np.dot(metric.matlap, temp0.ravel()).reshape(temp0.shape)

    temp0_1d = np.ones(metric.shp1d) * params["t_cold"]
    temp0_2d = temp0_1d.reshape(metric.shp2d)

    if params["method"] == "iterative":
        mat_bc_csr, rhs = apply_bc_csr(
            metric.lapl,
            metric.size_i,
            metric.size_j,
            params["t_cold"],
            params["t_hot"])

        temp1d, info = scp.linalg.bicgstab(
            mat_bc_csr,
            rhs,
            x0=temp0_1d,
            tol=1e-1
        )
        print(">> Result bicg ", info)
        temp = temp1d.reshape(metric.shp2d)

    if params["method"] == "direct":
        temp = explicit_solve(
            temp0_2d,
            metric,
            params,
            const,
            temp_final)

    def gap(temp):
        """ return the maximum absolute error """
        return np.abs(temp_final-temp).max()
    print("Max error:", gap(temp))

    outh5 = NpArray2Xmf("out.h5")
    outh5.create_grid(x_coor,
                      y_coor,
                      np.zeros(metric.shp2d))
    outh5.add_field(temp, "temp")
    outh5.add_field(temp_final-temp, "temp_error")
    outh5.dump()


if __name__ == "__main__":

    PARAMS = {"res": 1000,
              "max_loop": 1000000,
              "dilate_grid": True,
              "method": "iterative",
              #"method": "direct",
              "tol": 1.0,
              "fourier": 0.8,
              "typ_size": 0.1,
              "width": 0.001,
              "rho": 7900.,
              "cp": 435.,
              "lambda": 11.4,
              "t_hot": 1500.,
              "t_cold": 600.,
              "h_hot": 1500.,
              "h_cold": 3000.}
    main_2dheattreansf(PARAMS)



"""interpolate a cloud from an other cloud """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy import spatial
from arnica.utils.vector_actions import (yz_to_theta)

SMALL = 1e-16

def _build_ravel_component(in_xyz, limitsize=None):
    """ build source and target raveled components

    Parameters:
    -----------
    in_xyz (dict of nparray): las index are coordinantes 3D. (:,:,3)
    limitsize : maximum size of raveled output

    Returns:
    --------
    out_xyz : raveld version, limited in size
    """
    skipper = 1
    if limitsize is not None:
        input_size = np.ravel(np.take(in_xyz, 0, axis=-1)).shape[0]
        skipper = max(int(input_size * 1.0 / limitsize), 1)
        if skipper > 1:
            print("Warning, source data too big ({0}/{1}),\n "
                  "sub sampling every {2}th element.".format(input_size,
                                                             limitsize,
                                                             skipper))
    out_xyz = np.stack((np.ravel(np.take(in_xyz, 0, axis=-1))[::skipper],
                        np.ravel(np.take(in_xyz, 1, axis=-1))[::skipper],
                        np.ravel(np.take(in_xyz, 2, axis=-1))[::skipper]),
                       axis=1
                      )

    return out_xyz, skipper


def _interpolate_var(in_val, index, inv_dist):
    """interpolate nP_array of source
    intour a np.arry of target"""

    estimate = (np.sum(inv_dist *
                       np.reshape(in_val[np.ravel(index)],
                                  np.shape(inv_dist)),
                       axis=1))

    estimate /= np.sum(inv_dist, axis=1)

    return estimate



def cloud2cloud(source_xyz,
                source_val,
                target_xyz,
                stencil=1,
                limitsource=None,
                power=1.0,
                tol=None,
                #alter_metric="sphere",
                ):
    """ Interpolate  form a cloud to an other

    Parameters :
    ------------
    source_xyz : numpy array shape (n_s, 3) either (1000, 3 )  or (10,10,10, 3)
    source_val : numpy array shape (n_s, k) of k variables
    target_xyz : numpy array shape (n_t, 3)

    stencil (int): nb of neigbors to compute (1 is closest point)
    limitsource (int) : maximum nb of source points allowed (subsample beyond)

    Returns :
    ----------
    target_val : numpy array shape (n_t, k)

    """
    sce_ravel, skipper = _build_ravel_component(source_xyz,
                                                limitsize=limitsource)
    tgt_ravel, _ = _build_ravel_component(target_xyz)


    kdtree = spatial.cKDTree(sce_ravel)

    dists, index = kdtree.query(tgt_ravel, k=stencil,)

    # if alter_metric == "cubic":
    #     dists = np.amax(np.abs(tgt_ravel[:, np.newaxis, :]
    #                               - sce_ravel[index, :]),
    #                     axis=-1)
    #     print "cubic estimation"
    # 
    # if alter_metric == "cubic_cyl":
    #     x_dif = np.abs(sce_ravel[index, 0]
    #                    - tgt_ravel[:, np.newaxis, 0])
    #     t_dif = np.abs(yz_to_theta(sce_ravel[index, :])
    #                    - yz_to_theta(tgt_ravel)[:, np.newaxis])
    #     r_tgt = np.hypot(tgt_ravel[:, 1],
    #                      tgt_ravel[:, 2])
    #     r_dif = np.abs((np.hypot(sce_ravel[index, 1],
    #                              sce_ravel[index, 2])
    #                     - r_tgt[:, np.newaxis]))
    #     dists =  np.maximum(r_tgt[:, np.newaxis]*t_dif,
    #                         np.maximum(x_dif, r_dif))

    if power != 1.0:
        dists = np.power(dists, power)

    inv_dist = np.reciprocal(np.maximum(dists, SMALL))

    target_val = {}
    for key in source_val:
        if stencil > 1:
            estimate = _interpolate_var(np.ravel(source_val[key])[::skipper],
                                        index,
                                        inv_dist,
                                        )
        else:
            estimate = np.ravel(source_val[key])[::skipper][index]

        if tol is not None:
            estimate = np.where(dists[:, 0] > tol, 0, estimate)

        target_val[key] = np.reshape(estimate,
                                     np.take(target_xyz, 0, axis=-1).shape)




    return target_val



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='arnica',
    version='1.0.4a',
    description='Open Source library CFD toolkit',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='CoopTeam-CERFACS',
    author_email='coop@cerfacs.com',
    url='',
    keywords=["ARNICA"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    license="CeCILL-B FREE SOFTWARE LICENSE AGREEMENT",
    packages=find_packages(exclude=('tests', 'docs')),
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    install_requires=[
        'numpy',
        'scipy',
        'sympy',
        'matplotlib',
        'pandas',
        'h5py',
        'lxml']
)



""" module to define a solid material for thermal computations """

from __future__ import absolute_import, division
import numpy as np
from matplotlib import pyplot as plt


class SolidMaterial(object):
    """ define properties of a solid material object """
    def __init__(self,
                 lambda_poly,
                 lambda_range):
        """ startup class"""
        self._lambda_range = lambda_range
        self._lambda_raw = np.poly1d(lambda_poly[::-1])

    def lambda_th(self, temperature):
        """ return the lambda of ceramics material [W/mK] """
        t_clip = np.minimum(temperature, self._lambda_range[1])
        t_clip = np.maximum(t_clip, self._lambda_range[0])
        return self._lambda_raw(t_clip)

    def thermal_resistance(self, width, t_est):
        """ return the thermal resistance  [m2.K/W]
        width : width of the layer
        t_est : estimated temperature of the layer """
        return width / self.lambda_th(t_est)
        



""" axishell module
to creat axy symmetric shells for scientific computations
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json

import numpy as np
from scipy import interpolate, spatial

MESH_CTRL_POINTS = "axishell.json"
MAX_SPLINE_ORDER = 3
SPLINE_SMOOTHNESS = 0

MATRIX_ELEMENTS = ["xyz", "r", "theta",
                   "n_x", "n_y", "n_z", "n_r"]

AXIS_NAMES = ['time', 'theta', 'r']


class AxiShell():
    """ base clase for axysymmetric computationnal shells"""

    def __init__(self, n_longi, n_azi):
        """ startup method"""

        self.shape = (n_azi, n_longi)
        # dict for the definition of the shell
        self.geom = {}

        # numpy array for the node based geometric elements
        self.matrix = {}

        self.cake = {}

        # a matrix registering different width curvilinear based.
        self.matrix_widths = {}

    def init_mockup(self):
        """ initialize with a mockup mesh"""
        self.geom["angle"] = 40.0

        # by nodes trapvtx
        self.geom["crtl_pts_x"] = (0.0,
                                   0.2)
        self.geom["crtl_pts_r"] = (0.12,
                                   0.12)

        # # sphere
        # self.geom["crtl_pts_x"] = 0.1 + 0.1* np.sin(
        #     np.linspace(-np.pi*0.5, np.pi*0.5, 4))
        # self.geom["crtl_pts_r"] = 0.1*np.cos(
        #     np.linspace(-np.pi*0.5, np.pi*0.5, 4))

        # by nodes curved shape
        # self.geom["crtl_pts_x"] = (0.0,
        #                            0.1,
        #                            0.2,
        #                            0.15)
        # self.geom["crtl_pts_r"] = (0.1,
        #                            0.15,
        #                            0.1,
        #                            0.03)

        self.build_shell()

    def load(self):
        """ load axishell from json"""
        with open(MESH_CTRL_POINTS, "w") as fin:
            self.geom = json.load(fin)

        self.build_shell()

    def dump(self):
        """ dump axishell from json"""
        with open(MESH_CTRL_POINTS, "w") as fout:
            json.dump(self.geom, fout)

    def build_shell(self):
        """ build shell from geometry """

        # Build crest
        spline_order = min(len(self.geom["crtl_pts_x"]) - 1, MAX_SPLINE_ORDER)
        tck, _ = interpolate.splprep([self.geom["crtl_pts_x"],
                                      self.geom["crtl_pts_r"]],
                                     s=SPLINE_SMOOTHNESS,
                                     k=spline_order)
        unew = np.linspace(0,
                           1,
                           num=self.shape[1])
        shell_crest = np.asarray(interpolate.splev(unew, tck))

        normal_vects = np.roll(np.diff(shell_crest), 1, axis=0)
        normal_vects /= np.linalg.norm(normal_vects, axis=0)
        normal_vects[0, :] *= -1
        normal_vects2 = np.ones((2, self.shape[1]))
        normal_vects2[:, :self.shape[1] - 1] = normal_vects
        normal_vects2[:, -1] = normal_vects[:, -1]

        for item in MATRIX_ELEMENTS:
            self.matrix[item] = np.ones(self.shape)

        tmp_x = np.tile(shell_crest[0], (self.shape[0], 1))
        self.matrix["r"] = np.tile(shell_crest[1], (self.shape[0], 1))
        self.matrix["n_x"] = np.tile(normal_vects2[0], (self.shape[0], 1))
        self.matrix["n_r"] = np.tile(normal_vects2[1], (self.shape[0], 1))
        self.matrix["theta"] = np.transpose(
            np.tile(np.linspace(-0.5 * self.geom["angle"] * np.pi / 180,
                                0.5 * self.geom["angle"] * np.pi / 180,
                                num=self.shape[0]),
                    (self.shape[1], 1)))

        tmp_y = (self.matrix["r"] * np.cos(self.matrix["theta"]))
        tmp_z = (self.matrix["r"] * np.sin(self.matrix["theta"]))

        self.matrix["xyz"] = np.stack((tmp_x, tmp_y, tmp_z),
                                      axis=2)

        self.matrix["n_y"] = (self.matrix["n_r"]
                              * np.cos(self.matrix["theta"]))
        self.matrix["n_z"] = (self.matrix["n_r"]
                              * np.sin(self.matrix["theta"]))

    def width_mockup(self):
        """ create a mockup tuple of tuple for widths"""

        width_mockup = ((0.0, -0.04),
                        (0.2, -0.04),
                        (0.3, -0.07),
                        (0.4, -0.07),
                        (0.5, -0.04),
                        (0.8, -0.04))

        return width_mockup

    def add_curviwidth(self, title, data):
        """ add a curvilinear with fiel to shell """

        (x_tuple, y_tuple) = tuple(zip(*data))

        # extend bounds
        xlist = [0] + list(x_tuple) + [1]
        ylist = [y_tuple[0]] + list(y_tuple) + [y_tuple[-1]]
        f_int = interpolate.interp1d(xlist, ylist)
        xnew = np.linspace(0, 1, num=self.shape[1])
        ynew = f_int(xnew)

        self.matrix[title] = np.tile(ynew, (self.shape[0], 1))

    def set_mask_on_shell(self, cloud, tol):
        """ create a mask on the shell from a cloud
        """
        kdtree = spatial.KDTree(cloud)
        dists, _ = kdtree.query(self.matrix["xyz"], k=1)
        self.matrix["mask"] = np.where(
            dists.reshape(self.shape) > tol,
            np.zeros(self.shape),
            np.ones(self.shape))

    def bake_millefeuille(self, mwidth, n_layers, shift=0):
        """Create a milefeuille.

        extruding a shell
        in the normal direction
        scaled by matrix 'mwidth'
        on 'nlayers"
        shift : additionall depth
        "Bon appetit!"
        """
        cake = {}
        cake["xyz"] = np.empty((self.shape[0],
                                self.shape[1],
                                n_layers,
                                3))
        cake["dz"] = np.empty((self.shape[0],
                               self.shape[1],
                               n_layers))

        for j, dim in enumerate(["x", "y", "z"]):
            for i in range(n_layers):
                cake["xyz"][:, :, i, j] = (self.matrix["xyz"][:, :, j]
                                           + (1.0 * i / n_layers
                                              * self.matrix[mwidth] +
                                              shift)
                                           * self.matrix["n_" + dim])

        for i in range(n_layers):
            cake["dz"][:, :, i] = abs(self.matrix[mwidth] / n_layers)
        cake["dz"][:, :, 0] /= 2
        cake["dz"][:, :, -1] /= 2
        return cake


    def average_on_shell_over_dirs(self, variable, directions):
        """Performs an integration (averaging) over one or
            multiple directions : time, theta and r


            Parameters:
            ==========
            variable  : the array to be averaged of shape
                        (n_time, n_theta, n_r)
            directions : list of directions on which average is
                         to be performed should contain keywords from
                         ['time','theta', 'r']

            Returns:
            =======
            averaged_variable : integrated(averaged) array on the given
                                directions
        """

        scaled = np.copy(variable)
        if self.matrix.get("surf") is not None:
            scaled = np.multiply(variable, self.matrix.get("surf"))

        dirs = AXIS_NAMES.copy()
        cond = directions and len(directions) <= 3
        cond = cond and all(d in dirs for d in directions)
        if not cond:
            mess = 'Averaging directions do not conform to criteria\n'
            mess = mess +'It should be a list containing one of or all items\n'
            mess = mess +'"time", "theta", "r"'
            raise ValueError(mess)
        else:
            averaged_variable = scaled
            for dir_ in directions:
                index = dirs.index(dir_)
                averaged_variable = np.mean(averaged_variable, axis=index)
                dirs.pop(index)
        return averaged_variable
